from bs4.element import Tag

import pybloqs
from pybloqs.server import BloqsProvider
from pybloqs.server.static import HTMX


class Poll(pybloqs.BaseBlock):
    resource_deps = (HTMX,)

    def __init__(self, provider: BloqsProvider, frequency: str = "10s", *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.provider = provider
        self.frequency = frequency

    def _write_contents(self, container: Tag, actual_cfg, id_gen, resource_deps=None, static_output=None) -> None:
        container["hx-get"] = self.provider.url
        container["hx-trigger"] = f"load, every {self.frequency}"
        container["hx-swap"] = "innerHTML"



from typing import Any, Dict, Iterable, Optional, Type, Union

from pybloqs.block.base import BaseBlock
from pybloqs.block.text import Raw

_block_types: Dict[Type, Type[BaseBlock]] = {}


def add_block_types(objects: Union[Type, Iterable[Type]], block_cls: Type[BaseBlock]) -> None:
    if not isinstance(objects, Iterable):
        objects = [objects]
    for o in objects:
        _block_types[o] = block_cls


# noinspection PyPep8Naming
def Block(
    contents: Any = None,
    title: Optional[str] = None,
    title_level: int = 3,
    title_wrap: bool = False,
    inherit_cfg: bool = True,
    **kwargs,
) -> BaseBlock:
    """
    Constructs a composable layout element that will be rendered automatically by
    IPython Notebooks. It can also be saved in HTML, PDF, PNG or JPEG formats.

    Content is handled as follows:

    - Lists, tuples and sets are written out into a grid layout, with a single column
      by default. Individual elements of the grid are parsed recursively.
    - DataFrames are written out as an interactive HTML table.
    - Strings are written out in a raw format, preserving any HTML content in them.
    - Nested blocks are simply wrapped, unless there is more than one in which case
      the same logic applies as for lists, tuples and sets.

    :param contents: Contents to put in a block.
    :param title: Optional title of the block.
    :param title_level: Optional title level (adjusts the size and TOC level), 1 being the
                        biggest and 9 being the smallest.
    :param title_wrap: Toggles whitespace wrapping of the title. (Default: False).
    :param inherit_cfg: Optional. Set to False to ensure that the block does not inherit
                        any parameters from parents.
    :param cascade_cfg: Set to True to enable parmater cascading from this block. A value
                        of False means that child blocks do not inherit parameters from
                        this block.
    :param kwargs: Optional styling arguments. The `style` keyword argument has special
                   meaning in that it allows styling to be grouped as one argument.
                   It is also useful in case a styling parameter name clashes with a standard
                   block parameter.
    :return: A block instance.
    """
    block_cls = None
    # Need to loop to catch inherited classes as well
    for key, value in _block_types.items():
        if isinstance(contents, key):
            block_cls = value
            break

    # Try some additional transformations if no suitable mapping found
    if block_cls is None:
        if isinstance(contents, str):
            block_cls = Raw
        elif isinstance(contents, BaseBlock):
            # If there is no title, there is no point to wrap the existing block
            if title is None:
                return contents

            class _NestedBlock(BaseBlock):
                container_tag = None

                def __init__(self) -> None:
                    super().__init__(
                        title=title, title_level=title_level, title_wrap=title_wrap, inherit_cfg=inherit_cfg, **kwargs
                    )

                def _to_static(self) -> BaseBlock:
                    return contents._to_static()

                def _write_contents(self, *sub_args, **sub_kwargs) -> None:
                    contents._write_block(*sub_args, **sub_kwargs)

            return _NestedBlock()
        elif contents is None:
            block_cls = Raw
            contents = ""
        else:
            raise ValueError(f"Unrecognized argument type: {type(contents)}")

    return block_cls(
        contents, title=title, title_level=title_level, title_wrap=title_wrap, inherit_cfg=inherit_cfg, **kwargs
    )



try:
    import altair as alt
    from altair.vegalite.api import Chart as AltairChart

    _ALTAIR_AVAILABLE = True
except ImportError:
    _ALTAIR_AVAILABLE = False

from pybloqs.block.base import BaseBlock
from pybloqs.block.convenience import add_block_types
from pybloqs.html import parse


class VegaAltairBlock(BaseBlock):
    def __init__(self, contents, **kwargs) -> None:
        """
        A Block that renders Vega-Altair charts
        """
        if not isinstance(contents, alt.Chart):
            raise ValueError("Expected alt.Chart type but got %s", type(contents))

        super().__init__(**kwargs)
        self._fig = contents

    def _write_contents(self, container, *args, **kwargs) -> None:
        container.append(parse(self._fig.to_html(fullhtml=False)))


if _ALTAIR_AVAILABLE:
    add_block_types(AltairChart, VegaAltairBlock)



import inspect
import urllib.parse
from copy import deepcopy
from functools import partial
from io import BytesIO
from typing import Callable, Optional

from bs4.element import Tag
from flask import request

import pybloqs
from pybloqs.html import append_to, id_generator, render, root
from pybloqs.server.static import HTMX
from pybloqs.static import DependencyTracker
from pybloqs.util import Cfg


# This inherits from BaseBlock as it makes sense to provide it in a static
# block if the function has no parameters. You may want to do this, for example
# if you have a function that returns different blocks but want to use `partial`.
class BloqsProvider(pybloqs.BaseBlock):
    resource_deps = (HTMX,)

    def __init__(
        self,
        inner: Callable[..., pybloqs.BaseBlock],
        loading_block: Optional[pybloqs.BaseBlock] = None,
        **kwargs,
    ) -> None:
        """
        Parameters
        ----------
        kwargs: Passed through to constructor of BaseBlock
        """
        from pybloqs.server.block.life_loading import LifeLoadingBlock

        super().__init__(**kwargs)
        self.url = f"/{self._id}"
        self.inner = inner
        self.loading_block = loading_block or LifeLoadingBlock()

    def with_parameters(self, *args, **kwargs) -> "BloqsProvider":
        provider = deepcopy(self)
        url_parts = urllib.parse.urlparse(self.url)
        query = urllib.parse.parse_qs(url_parts.query)
        if args:
            query["args"] = [str(x) for x in args]
        query.update(kwargs)
        url_parts = url_parts._replace(query=urllib.parse.urlencode(query, True))
        provider.url = urllib.parse.urlunparse(url_parts)
        provider.inner = partial(self.inner, *args, **kwargs)
        return provider

    with_parameter = with_parameters

    def get_fragment(self, *args, **kwargs) -> str:
        already_sent_resources = []
        if not args and not kwargs:
            request_args = request.args.to_dict()
            args = request_args.pop("args", [])
            if not isinstance(args, list):
                args = [args]
            kwargs = request_args
            already_sent_resources = request.headers.get("Blox-Resources", "").split(",")
        block = self(*args, **kwargs)
        resource_deps = DependencyTracker()
        container = root("div")
        block._write_block(container, Cfg(), id_generator(), resource_deps)

        # Write children into the output
        output = BytesIO()

        for child in container.children:
            output.write(render(child).encode("utf-8"))

        container = root("div")
        container["hx-swap-oob"] = "beforeend:head"
        resources_added = []
        for dependency in resource_deps:
            if dependency.name not in already_sent_resources:
                dependency.write(container)
                resources_added.append(dependency.name)
        if resources_added:
            sentinel_js = append_to(container, "script")
            sentinel_js["type"] = "text/javascript"
            sentinel_js.string = f"""
                headers = JSON.parse(
                  document
                    .querySelector('body')
                    .getAttribute('hx-headers') || "{{}}"
                );
                headers['Blox-Resources'] =
                  (headers['Blox-Resources'] || "")
                  + ",{','.join(resources_added)}";
                document
                  .querySelector('body')
                  .setAttribute(
                    'hx-headers',
                    JSON.stringify(headers)
                  );
            """
        output.write(render(container).encode("utf-8"))

        return output.getvalue().decode()

    def __call__(self, *args, **kwargs) -> pybloqs.BaseBlock:
        """
        A provider acts just like the function it wraps.
        """
        return self.inner(*args, **kwargs)

    def _write_block(self, parent: Tag, parent_cfg, id_gen, resource_deps=None, static_output=False) -> None:
        """
        If this provider has no unbound parameters, render it as a lazy PyBlock.
        """
        unbound_parameters = [
            parameter
            for parameter in inspect.signature(self.inner).parameters.values()
            if parameter.default == inspect.Parameter.empty
        ]

        if not unbound_parameters:
            if resource_deps is not None:
                for res in self.resource_deps:
                    resource_deps.add(res)
            div_to_replace = append_to(parent, "div", **{"hx-trigger": "revealed", "hx-get": self.url})
            if self.loading_block:
                self.loading_block._write_block(
                    div_to_replace, parent_cfg, id_gen, resource_deps=resource_deps, static_output=static_output
                )
        else:
            raise ValueError(f"Cannot write block with missing parameters {unbound_parameters} {self.inner}")

    def poll(self, frequency: str = "10s") -> pybloqs.BaseBlock:
        from pybloqs.server.block import Poll

        return Poll(self, frequency)



from pybloqs.server.block.poll import Poll  # noqa: F401
from pybloqs.server.block.refresh import Refresh  # noqa: F401
from pybloqs.server.block.select import Select  # noqa: F401
from pybloqs.server.block.tabs import Tabs  # noqa: F401



import base64
import calendar
import datetime
import itertools
import re
import zlib
from typing import Any, Callable, Dict, Iterator, Tuple, Union

import numpy as np

try:
    from importlib.resources import files

    def get_resource_path(package, resource_name) -> str:
        """
        Workaround for 3.8 - 3.12+ compatibility. Returns the resource path

        :param package: The package name (pass __package__)
        :param resource_name: The name of the resource (file name)
        :return: Full resource path of the file
        """
        return str(files(package).joinpath(resource_name))

except ImportError:
    from pkg_resources import resource_filename

    def get_resource_path(package, resource_name) -> str:
        """
        Workaround for 3.8 - 3.12+ compatibility. Returns the resource path

        :param package: The package name (pass __package__)
        :param resource_name: The name of the resource (file name)
        :return: Full resource path of the file
        """
        return resource_filename(package, resource_name)


def dt_epoch_msecs(value: Union[datetime.date, datetime.datetime]) -> float:
    """
    Calculate miliseconds since epoch start for python datetimes.
    """
    return int(calendar.timegm(value.timetuple())) * 1000


def np_dt_epoch_msec(value: np.datetime64) -> float:
    """
    Calculate miliseconds since epoch start for numpy datetimes.
    """
    return value.astype(int) / 1000


def encode_string(string: str, level: int = 9) -> bytes:
    """
    Compresses and base64 encodes the supplied string

    :param string: String to encode and compress
    :param level: Compression level
    :return: Compressed and encoded string
    """
    return base64.b64encode(zlib.compress(string.encode("utf-8"), level))


def camelcase(value: str) -> str:
    """
    Converts 'under_score_string' -> 'underScoreString'

    :param value: Underscore separated string
    :return: CamelCased string
    """
    rest = value.split("_")
    return rest[0] + "".join(word.title() for word in rest[1:])


def underscorecase(camelcased: str) -> str:
    """
    Converts 'underScoreString' -> 'under_score_string'

    :param value: CamelCased string
    :return: Underscore separated string
    """
    return re.sub("([A-Z]+)", r"_\1", camelcased).lower()


def cfg_to_prop_string(
    cfg: "Cfg",
    key_transform: Callable[[str], str] = lambda k: k,
    value_transform: Callable[[str], str] = lambda v: v,
    separator: str = ";",
) -> str:
    """
    Convert the config object to a property string. Useful for constructing CSS and javascript
    object init strings.

    Underscores are replaced with dashes and values are converted to lower case.
    """
    return separator.join([f"{key_transform(key)}:{value_transform(value)}" for key, value in cfg.items()])


def cfg_to_css_string(cfg: "Cfg") -> str:
    return cfg_to_prop_string(cfg, lambda k: k.replace("_", "-"), lambda v: str(v).lower())


class Cfg(dict):
    """
    A dict-like mapping for inheritable block parameters.
    """

    def inherit(self, parent: Union[dict, "Cfg"]) -> "Cfg":
        """
        Inherit all parent settings that the current config does not define.
        """
        return self.__class__(Cfg._mergedicts(self, parent, False))

    def inherit_many(self, *args: "Cfg", **kwargs: Any) -> "Cfg":
        """
        Inherit many settings at once.
        """
        inherit_collapsed = self._collapse_args(args, kwargs)
        return self.inherit(inherit_collapsed)

    def override(self, parent: Union[Dict, "Cfg"]) -> "Cfg":
        """
        Override all settings specified in the overrides.
        """
        return self.__class__(Cfg._mergedicts(self, parent, True))

    def override_many(self, *args: "Cfg", **kwargs: Any) -> "Cfg":
        """
        Override many settings at once.
        """
        override_collapsed = self._collapse_args(args, kwargs)
        return self.override(override_collapsed)

    @staticmethod
    def _collapse_args(args: Tuple["Cfg", ...], kwargs: Any) -> "Cfg":
        inherit_agg = Cfg(kwargs)
        for arg in args:
            # If we got a type as configuration, instantiate it
            if isinstance(arg, type):
                arg = arg()

            inherit_agg = inherit_agg.inherit(arg)
        return inherit_agg

    @staticmethod
    def _mergedicts(dict1: Dict, dict2: Dict, take_second: bool) -> Iterator:
        for k in set(itertools.chain(dict1.keys(), dict2.keys())):
            if k in dict1 and k in dict2:
                v1 = dict1[k]
                v2 = dict2[k]
                if isinstance(v1, dict) and isinstance(v2, dict):
                    # The merged sub-dict type should be the operand type based on take_second
                    ctor = v2.__class__ if take_second else v1.__class__
                    yield (k, ctor(Cfg._mergedicts(v1, v2, take_second)))
                else:
                    yield (k, v2 if take_second else v1)
            elif k in dict1:
                yield (k, dict1[k])
            else:
                yield (k, dict2[k])

    def __getattr__(self, item: str) -> Any:
        try:
            return self[item]
        except KeyError:
            raise AttributeError(f"'Cfg' object has no attribute '{item}'")

    def __setattr__(self, name: str, value: Any) -> None:
        self[name] = value

    def __setstate__(self, state) -> None:
        for key in state:
            self[key] = state[key]

    def __getstate__(self) -> Dict[str, Any]:
        return {key: self[key] for key in self}



import logging
import os
import subprocess
import sys
from abc import ABCMeta, abstractmethod
from numbers import Number
from typing import TYPE_CHECKING, Collection, Tuple, Union

from pybloqs.config import user_config

if TYPE_CHECKING:
    from pybloqs import BaseBlock

logger = logging.getLogger(__name__)

PORTRAIT = "Portrait"
LANDSCAPE = "Landscape"
A4 = "A4"


class HTMLConverter:
    """
    Definition of interface for HTML to X converters
    """

    __metaclass__ = ABCMeta

    def get_executable(self, command_name: str) -> str:
        """Look for executable file name in VENV_NAME/bin/ otherwise use plain command name and hope it is in PATH."""
        py_bin_dir = os.path.join(sys.exec_prefix, "bin")
        local_bin = os.path.join(py_bin_dir, command_name)
        if os.path.isfile(local_bin):
            command = local_bin
        else:
            command = command_name
        return command

    def run_command(self, cmd: str) -> Tuple[bytes, bytes]:
        """Run cmd as subprocess and return stdout and stderr."""
        logger.info("Running external application: %s", cmd)
        proc = subprocess.Popen(cmd)

        # Wait for the process to exit
        output, errors = proc.communicate()

        if proc.returncode != 0:
            raise ValueError(
                f"{cmd} returned:\n stdout:{output}\n stderr:{errors}",
            )
        else:
            logger.info("Returned:\n stdout: %s\n stderr:%s", output, errors)
        return output, errors

    @staticmethod
    def write_html_to_tempfile(block: "BaseBlock", content: str) -> str:
        name = block._id[: user_config["id_precision"]] + ".html"
        tempdir = user_config["tmp_html_dir"]
        html_filename = os.path.join(tempdir, name)
        with open(html_filename, "w", encoding="utf-8") as f:
            f.write(content)
        return html_filename

    @staticmethod
    def remove_temporary_files(temp_files: Collection[str]) -> None:
        if user_config.get("remove_temp_files", True):
            for f in temp_files:
                try:
                    os.remove(f)
                except OSError:
                    logger.exception("Failed to remove a temporary file: %s.", f)

    @abstractmethod
    def htmlconv(
        self,
        block: "BaseBlock",
        output_file: str,
        header_filename=None,
        header_spacing: Union[None, str, Number] = None,
        footer_filename=None,
        footer_spacing: Union[None, str, Number] = None,
        zoom: Number = 1,
        pdf_page_size: str = A4,
        orientation: str = PORTRAIT,
        **kwargs,
    ) -> Tuple[bytes, bytes]:
        pass



import contextlib
import os
import uuid
import webbrowser
from io import BytesIO
from typing import Any, Iterable, Iterator, Optional, Tuple, Union
from urllib.parse import urljoin

import pybloqs.htmlconv as htmlconv
from pybloqs.config import user_config
from pybloqs.email import send_html_report
from pybloqs.html import append_to, id_generator, js_elem, render, root
from pybloqs.static import Css, DependencyTracker, register_interactive, script_block_core, script_inflate
from pybloqs.util import Cfg, cfg_to_css_string

default_css_main = Css(os.path.join("css", "pybloqs_default", "main"))
register_interactive(default_css_main)


class BaseBlock:
    """
    Base class for all blocks. Provides infrastructure for rendering the block
    in an IPython Notebook or saving it to disk in HTML, PDF, PNG or JPG format.
    """

    container_tag = "div"
    resource_deps = ()

    def __init__(
        self,
        title: Optional[str] = None,
        title_level: int = 3,
        title_wrap: bool = False,
        width=None,
        height=None,
        inherit_cfg: bool = True,
        styles=None,
        classes: Union[str, Iterable[str]] = (),
        anchor=None,
        id_: Optional[str] = None,
        **kwargs,
    ) -> None:
        self._settings = Cfg(
            title=title,
            title_level=title_level,
            title_wrap=title_wrap,
            cascading_cfg=Cfg(**kwargs).override(styles or Cfg()),
            default_cfg=Cfg(),
            inherit_cfg=inherit_cfg,
            width=width,
            height=height,
            classes=["pybloqs"] + ([classes] if isinstance(classes, str) else list(classes)),
        )
        # Anchor should not be inherited, so keep outside of Cfg
        self._anchor = anchor
        self._id = id_ or uuid.uuid4().hex

    def render_html(
        self,
        pretty: bool = True,
        static_output: bool = False,
        header_block: Optional["BaseBlock"] = None,
        footer_block: Optional["BaseBlock"] = None,
        permit_compression: bool = True,
    ) -> str:
        """
        Returns html output of the block

        :param pretty: Toggles pretty printing of the resulting HTML. Not applicable for non-HTML output.
        :param static_output: Passed down to _write_block. Will render static version of blocks which support this.
        :param header_block: If not None, header_block is inlined into a HTML body as table.
        :param footer_block: If not None, footer_block is inlined into a HTML body as table.
        :param permit_compression: If set, resources will be embedded as base64 gzipped files
        :return html-code of the block
        """
        # Render the contents
        html = root("html", doctype="html")
        head = append_to(html, "head")
        append_to(head, "meta", charset="utf-8")

        body = append_to(html, "body")

        # Make sure that the main style sheet is always included
        resource_deps = DependencyTracker(default_css_main)

        # If header or footer are passed into this function, inline them in the following structure:
        #
        # <body>
        # <table>
        #    <thead><tr><td>Header html</td></tr></thead>
        #    <tfoot><tr><td>Footer html</td></tr></tfoot>
        #    <tbody><tr><td>Body html</td></tr></tbody>
        # </table>
        # </body>
        if header_block is not None or footer_block is not None:
            content_table = append_to(body, "table")
            if header_block is not None:
                header_thead = append_to(content_table, "thead")
                header_tr = append_to(header_thead, "tr")
                header_td = append_to(header_tr, "th")
                header_block._write_block(
                    header_td, Cfg(), id_generator(), resource_deps=resource_deps, static_output=static_output
                )

            if footer_block is not None:
                footer_tfoot = append_to(content_table, "tfoot", id="footer")
                footer_tr = append_to(footer_tfoot, "tr")
                footer_td = append_to(footer_tr, "td")
                footer_block._write_block(
                    footer_td, Cfg(), id_generator(), resource_deps=resource_deps, static_output=static_output
                )

            body_tbody = append_to(content_table, "tbody")
            body_tr = append_to(body_tbody, "tr")
            body_td = append_to(body_tr, "td")
            self._write_block(body_td, Cfg(), id_generator(), resource_deps=resource_deps, static_output=static_output)
        else:
            self._write_block(body, Cfg(), id_generator(), resource_deps=resource_deps, static_output=static_output)

        if permit_compression:
            script_inflate.write(head)
        script_block_core.write(head)

        if static_output:
            # Add the load wait poller if there are any JS resources
            js_elem(body, "var loadWaitPoller=runWaitPoller();")

        # Write out resources
        for res in resource_deps:
            res.write(head, permit_compression=permit_compression)

        # Render the whole document (the parent of the html tag)
        content = render(html.parent, pretty=pretty)
        return content

    def save(
        self,
        filename: Optional[str] = None,
        fmt: Optional[str] = None,
        pdf_zoom: float = 1,
        pdf_page_size: str = htmlconv.html_converter.A4,
        pdf_auto_shrink: bool = True,
        orientation: str = htmlconv.html_converter.PORTRAIT,
        header_block: Optional["BaseBlock"] = None,
        header_spacing: Union[str, float] = 5,
        footer_block: Optional["BaseBlock"] = None,
        footer_spacing: Union[str, float] = 5,
        **kwargs,
    ) -> str:
        """
        Render and save the block. Depending on whether the filename or the format is
        provided, the content will either be written out to a file or returned as a string.

        :param filename: Format will be based on the file extension.
                         The following formats are supported:
                         - HTML
                         - PDF
                         - PNG
                         - JPG
        :param fmt: Specifies the format of a temporary output file. When supplied, the filename
                    parameter must be omitted.
        :param pdf_zoom: The zooming to apply when rendering the page.
        :param pdf_page_size: The page size to use when rendering the page to PDF.
        :param pdf_auto_shrink: Toggles auto-shrinking content to fit the desired page size (wkhtmltopdf only)
        :param orientation: Either html_converter.PORTRAIT or html_converter.LANDSCAPE
        :param header_block: Block to be used as header (and repeated on every page). Only used for PDF output.
        :param header_spacing: Size of header block. Numbers are in mm. HTML sizes (e.g. '5cm') in chrome_headless only.
        :param footer_block: Block to be used as footer (and repeated on every page). Only used for PDF output.
        :param footer_spacing: Size of header block. Numbers are in mm. HTML sizes (e.g. '5cm') in chrome_headless only.
        :return: html filename
        """
        # Ensure that exactly one of filename or fmt is provided
        if filename is None and fmt is None:
            raise ValueError("One of `filename` or `fmt` must be provided.")

        tempdir = user_config["tmp_html_dir"]

        if filename:
            _, fmt_from_name = os.path.splitext(filename)
            # Exclude the dot from the extension, gosh darn it!
            fmt_from_name = fmt_from_name[1:]
            if fmt is None:
                if fmt_from_name == "":
                    raise ValueError("If fmt is not specified, filename must contain extension")
                fmt = fmt_from_name
            else:
                if fmt != fmt_from_name:
                    filename += "." + fmt
        else:
            name = self._id[: user_config["id_precision"]] + "." + fmt
            filename = os.path.join(tempdir, name)

        # Force extension to be lower case so format checks are easier later
        fmt = fmt.lower()

        is_html = "htm" in fmt

        if is_html:
            content = self.render_html(static_output=False, header_block=header_block, footer_block=footer_block)
            with open(filename, "w", encoding="utf-8") as f:
                f.write(content)
        else:
            converter = htmlconv.get_converter(fmt)
            converter.htmlconv(
                self,
                filename,
                header_block=header_block,
                header_spacing=header_spacing,
                footer_block=footer_block,
                footer_spacing=footer_spacing,
                pdf_page_size=pdf_page_size,
                orientation=orientation,
                pdf_auto_shrink=pdf_auto_shrink,
                pdf_zoom=pdf_zoom,
                **kwargs,
            )
        return filename

    def publish(self, name: str, *args, **kwargs) -> str:
        """
        Publish the block so that others can access it.

        :param name: Name to publish under. Can be a filename or a relative path.
        :param args: Arguments to pass to `Block.save`.
        :param kwargs: Keyword arguments to pass to `Block.save`.
        :return: Path to the published block file.
        """
        full_path = os.path.join(user_config["public_dir"], name)
        full_path = os.path.expanduser(full_path)

        base_dir = os.path.dirname(full_path)

        with contextlib.suppress(OSError):  # Ignore if directory already exists
            os.makedirs(base_dir)

        self.save(full_path, *args, **kwargs)

        return full_path

    def show(
        self, fmt: str = "html", header_block: Optional["BaseBlock"] = None, footer_block: Optional["BaseBlock"] = None
    ) -> str:
        """
        Show the block in a browser.

        :param fmt: The format of the saved block. Supports the same output as `Block.save`
        :return: Path to the block file.
        """
        file_name = self._id[: user_config["id_precision"]] + "." + fmt
        file_path = self.publish(
            os.path.expanduser(os.path.join(user_config["tmp_html_dir"], file_name)),
            header_block=header_block,
            footer_block=footer_block,
        )

        try:
            url_base = user_config["public_dir"]
        except KeyError:
            path = os.path.expanduser(file_path)
        else:
            path = urljoin(url_base, os.path.expanduser(user_config["tmp_html_dir"] + "/" + file_name))

        webbrowser.open_new_tab(path)

        return path

    def email(
        self,
        title: str = "",
        recipients: Tuple[str, ...] = (user_config["user_email_address"],),
        header_block: Optional["BaseBlock"] = None,
        footer_block: Optional["BaseBlock"] = None,
        from_address: Optional[str] = None,
        cc: Optional[str] = None,
        bcc: Optional[str] = None,
        attachments=None,
        convert_to_ascii: bool = True,
        **kwargs,
    ) -> None:
        """
        Send the rendered blocks as email. Each output format chosen will be added as an
        attachment.

        :param title: title of the email
        :param recipients: recipient of the email
        :param fmt: One or more output formats that should be included as attachments.
                    The following formats are supported:
                    - HTML
                    - PDF
                    - PNG
                    - JPG
        :param body_block: The block to use as the email body. The default behavior is
                          to use the current block.
        :param from_address: sender of the message. Defaults to user name.
            Can be overwritten in .pybloqs.cfg with yaml format: 'user_email_address: a@b.com'
        :param cc: cc recipient
        :param bcc: bcc recipient
        :param convert_to_ascii: bool to control convertion of html email to ascii or to leave in current format
        :param kwargs: Optional arguments to pass to `Block.render_html()`
        """
        if from_address is None:
            from_address = user_config["user_email_address"]

        # The email body needs to be static without any dynamic elements.
        email_html = self.render_html(header_block=header_block, footer_block=footer_block, **kwargs)

        send_html_report(
            email_html,
            recipients,
            subject=title,
            attachments=attachments,
            From=from_address,
            Cc=cc,
            Bcc=bcc,
            convert_to_ascii=convert_to_ascii,
        )

    def to_static(self) -> "BaseBlock":
        return self._visit(lambda block: block._to_static())

    def _to_static(self) -> "BaseBlock":
        """
        Subclasses can override this method to provide a static content version.
        """
        return self

    def _visit(self, visitor) -> Any:
        """
        Calls the supplied visitor function on this block and any sub-blocks
        :param visitor: Visitor function
        :return: Return value of the visitor
        """
        return visitor(self)

    def _provide_default_cfg(self, defaults) -> None:
        """
        Makes the supplied config to be part of the defaults for the block.
        :param defaults: The default parameters that should be inherited.
        """
        self._settings.default_cfg = self._settings.default_cfg.inherit(defaults)

    def _combine_parent_cfg(self, parent_cfg) -> Cfg:
        """from pybloqs.config import user_config
        Combine the supplied parent and the current Block's config.

        :param parent_cfg: Parent config to inherit from.
        :return: Combined config.
        """
        # Combine parameters only if inheritance is turned on
        if self._settings.inherit_cfg:
            actual_cfg = self._settings.cascading_cfg.inherit(parent_cfg)
        else:
            actual_cfg = self._settings.cascading_cfg

        # Any undefined settings will use the defaults
        actual_cfg = actual_cfg.inherit(self._settings.default_cfg)

        return actual_cfg

    def _get_styles_string(self, styles_cfg) -> str:
        """
        Converts the styles configuration to a CSS styles string.

        :param styles_cfg: The configuration object to convert.
        :return: CSS string
        """
        sizing_cfg = Cfg()

        if self._settings.width is not None:
            sizing_cfg["width"] = self._settings.width

        if self._settings.height is not None:
            sizing_cfg["height"] = self._settings.height

        # Replace `_` with `-` and make values lowercase to get valid CSS names
        return cfg_to_css_string(styles_cfg.override(sizing_cfg))

    def _write_block(self, parent, parent_cfg, id_gen, resource_deps=None, static_output: bool = False) -> None:
        """
        Writes out the block into the supplied stream, inheriting the parent_parameters.

        :param parent: Parent element
        :param parent_cfg: Parent parameters to inherit.
        :param id_gen: Unique ID generator.
        :param resource_deps: Object used to register resource dependencies.
        :param static_output: A value of True signals to blocks that the final output will
                              be a static format. Certain dynamic content will render with
                              alternate options.

        """
        if resource_deps is not None:
            for res in self.resource_deps:
                resource_deps.add(res)

        actual_cfg = self._combine_parent_cfg(parent_cfg)

        if self.container_tag is not None:
            container = append_to(parent, self.container_tag)
            self._write_container_attrs(container, actual_cfg)
        else:
            container = parent

        self._write_anchor(container)
        self._write_title(container)
        self._write_contents(container, actual_cfg, id_gen, resource_deps=resource_deps, static_output=static_output)

    def _write_container_attrs(self, container, actual_cfg) -> None:
        """
        Writes out the container attributes (styles, class, etc...).
        Note that this method will only be called if the container tag is not `None`.

        :param container: Container element.
        :param actual_cfg: Actual parameters to use.
        """
        styles = self._get_styles_string(actual_cfg)
        if len(styles) > 0:
            container["style"] = styles

        container["class"] = self._settings.classes

    def _write_title(self, container) -> None:
        """
        Write out the title (if there is any).

        :param container: Container element.
        """
        if self._settings.title is not None and (self._settings.title != ""):
            title = append_to(
                container,
                f"H{self._settings.title_level}",
                style="white-space: %s" % ("normal" if self._settings.title_wrap else "nowrap"),
            )
            title.string = self._settings.title

    def _write_anchor(self, container) -> None:
        """
        Write HTML anchor for linking within page

        :param container: Container element.
        """
        if self._anchor is not None:
            append_to(container, "a", name=self._anchor)

    def _write_contents(
        self, container, actual_cfg, id_gen: Iterator[str], resource_deps=None, static_output: Optional[bool] = None
    ) -> None:
        """
        Write out the actual contents of the block. Deriving classes must override
        this method.

        :param container: Container element.
        :param actual_cfg: Actual parameters to use.
        :param id_gen: Unique ID generator.
        :param resource_deps: Object used to register resource dependencies.
        :param static_output: A value of True signals to blocks that the final output will
                              be a static format. Certain dynamic content will render with
                              alternate options.
        """
        raise NotImplementedError("_write_contents")

    def _repr_html_(self, *_) -> str:
        """
        Function required to support interactive IPython plopping and plotting.

        Should not be used directly.

        :return: Data to be displayed
        """
        return self.data.decode()

    @property
    def data(self) -> bytes:
        """
        Function required to support interactive IPython plotting.

        Should not be used directly.

        :return: Data to be displayed
        """
        container = root("div")
        resource_deps = DependencyTracker(default_css_main)
        self._write_block(container, Cfg(), id_generator(), resource_deps=resource_deps)
        for resource in resource_deps:
            resource.write(container, permit_compression=False)

        # Write children into the output
        output = BytesIO()

        for child in container.children:
            output.write(render(child).encode("utf-8"))

        return output.getvalue()


class HRule(BaseBlock):
    """
    Draws a horizontal divider line.
    """

    def _write_block(self, parent, *_args, **_kwargs) -> None:
        # Add a `hr` element to the parent
        append_to(parent, "hr")



import json
from base64 import b64encode
from os import PathLike
from typing import Callable, Optional, Union
from uuid import uuid4

from flask import Flask

import pybloqs
from pybloqs.block.base import default_css_main
from pybloqs.html import append_to, id_generator, render, root
from pybloqs.server.provider import BloqsProvider
from pybloqs.static import DependencyTracker, script_block_core, script_inflate
from pybloqs.util import Cfg


def _getapp() -> Flask:
    if "app" not in globals():
        globals()["app"] = Flask(__name__)
    return globals()["app"]


def serve_block(
    base_block: pybloqs.BaseBlock,
    route: str = "/",
    title: Optional[str] = None,
    favicon: Union[PathLike, bytes, None] = None,
    logging_callback: Optional[Callable[[], None]] = None,
) -> None:
    # We re-implement pyblocks.BaseBlock.render_html so that we can inject a bunch of
    # stuff into the header. This could in theory be removed once pybloqs gets better.

    html = root("html", doctype="html")
    head = append_to(html, "head")
    if title:
        title_tag = append_to(head, "title", charset="utf-8")
        title_tag.string = title
    if favicon:
        if isinstance(favicon, bytes):
            favicon_data = favicon
        else:
            with open(favicon, "rb") as f:
                favicon_data = f.read()
        favicon_data_b64 = b64encode(favicon_data).decode()
        append_to(
            head,
            "link",
            rel="icon",
            href=f"data:image/x-icon;base64,{favicon_data_b64}",
        )

    body = append_to(html, "body")

    # Make sure that the main style sheet is always included
    resource_deps = DependencyTracker(default_css_main)

    base_block._write_block(
        body,
        Cfg(),
        id_generator(),
        resource_deps=resource_deps,
    )

    script_inflate.write(head)
    script_block_core.write(head)

    resources_added = []
    for res in resource_deps:
        res.write(head)
        resources_added.append(res.name)
    resources_str = ",".join(resources_added)
    body["hx-headers"] = json.dumps({"Blox-Resources": resources_str})
    # Render the whole document (the parent of the html tag)
    content = render(html.parent, pretty=True)

    def handler() -> str:
        if logging_callback:
            logging_callback()
        return content

    _getapp().add_url_rule(route, endpoint=str(uuid4()), view_func=handler)


def bloqs_provider(*args, **kwargs) -> BloqsProvider:
    is_bare = len(args) == 1 and not kwargs and callable(args[0])
    if is_bare:
        f = args[0]
        args = ()
        kwargs = {}
    elif args:
        raise ValueError(f"Cannot call bloqs_provider with arguments {args}. Use kwargs instead.")

    def wrapper(f: Callable) -> BloqsProvider:
        provider = BloqsProvider(f, *args, **kwargs)
        _getapp().add_url_rule(provider.url, endpoint=provider._id, view_func=provider.get_fragment)
        return provider

    if is_bare:
        return wrapper(f)
    return wrapper



from numbers import Number
from typing import TYPE_CHECKING, Optional, Tuple, Union

from pybloqs.htmlconv.html_converter import A4, PORTRAIT, HTMLConverter
from pybloqs.util import get_resource_path

if TYPE_CHECKING:
    from pybloqs import BaseBlock

_NODE_SCRIPT_LOC = get_resource_path(__package__, "puppeteer.js")


class ChromeHeadlessConverter(HTMLConverter):
    """
    Implementation of HTML to PDF conversion using headless chrome,
    accessed via puppeteer node.js bindings.

    See package.json at the root of this project for the prerequisite node.js
    dependencies
    """

    def htmlconv(
        self,
        block: "BaseBlock",
        output_file: str,
        header_block: Optional["BaseBlock"] = None,
        header_spacing: Union[None, str, Number] = None,
        footer_block: Optional["BaseBlock"] = None,
        footer_spacing: Union[None, str, Number] = None,
        pdf_zoom: Number = 1,
        pdf_page_size: str = A4,
        orientation: str = PORTRAIT,
        **kwargs,
    ) -> Tuple[bytes, bytes]:
        """
        Use headless chrome to convert the supplied html content to a PDF file.

        :param block: The block to render
        :param output_file: File name for converter output.
        :param header_block: A header block to add
        :param header_spacing: The spacing for the header, either HTML lenght unit or number (interpreted as mm).
        :param footer_block: A footer block to add
        :param footer_spacing: The spacing for the footer, either HTML lenght unit or number (interpreted as mm).
        :param pdf_zoom: The zooming to apply when rendering the page.
        :param pdf_page_size: The page size to use when rendering the page to PDF.
        :param orientation: Either html_converter.PORTRAIT or html_converter.LANDSCAPE
        :param kwargs: Additional parameters. Not currently used in chrome_headless backend.
        """
        # For compatibility with wkhtmltopdf handle spacing that is passed as number to be in mm.
        if header_spacing is not None and header_block is not None:
            if isinstance(header_spacing, Number):
                header_block._settings.height = f"{header_spacing}mm"
            else:
                header_block._settings.height = header_spacing
        if footer_spacing is not None and footer_block is not None:
            if isinstance(footer_spacing, Number):
                footer_block._settings.height = f"{footer_spacing}mm"
            else:
                footer_block._settings.height = footer_spacing
        content = block.render_html(static_output=True, header_block=header_block, footer_block=footer_block)

        html_filename = HTMLConverter.write_html_to_tempfile(block, content)
        output, errors = self.run_command(
            [
                self.get_executable("node"),
                _NODE_SCRIPT_LOC,
                "--zoom",
                str(pdf_zoom),
                "--landscape",
                "false" if orientation == PORTRAIT else "true",
                "--format",
                pdf_page_size,
                f"file://{html_filename}",
                output_file,
            ]
        )
        HTMLConverter.remove_temporary_files([html_filename])
        return output, errors



import base64
from typing import Iterator, Optional

import pybloqs
from pybloqs.html import append_to
from pybloqs.static import Css, JScript

BOX_SVG = """
<svg width="8mm" height="8mm" viewBox="0 0 7 7" version="1.1" id="svg5" xmlns="http://www.w3.org/2000/svg" xmlns:svg="http://www.w3.org/2000/svg">
  <rect
    style="fill:#000000;stroke-width:1.0;fill-opacity:0;stroke-miterlimit:4;stroke-dasharray:none;stroke:#C6D9F1;stroke-opacity:1"
    id="rec t84"
    width="7"
    height="7"
    x="0"
    y="0" />
</svg>
"""

CSS_STRING = f"""
.pybloqs_life {{
  position: relative;
  height: 200px;
  overflow: hidden;
}}
.pybloqs_life_background {{
  position: absolute;
  image-rendering: pixelated;
  top: 50%;
  left: 50%;
  width: 100%;
  height: 100%;
  transform: translate(-50%, -50%);
  display: flex;
  justify-content: center;
  align-items: center;
}}
.pybloqs_life_background > * {{
  position: absolute;
  flex-shrink: 0;
  object-fit: cover;
  height: 100%;
  width: 100%;
  aspect-ratio: 1;
}}
.pybloqs_life_background>.grid {{
  background-image: url('data:image/svg+xml;base64,{base64.b64encode(BOX_SVG.encode()).decode()}');
  background-size: 2%;
  height: unset;
}}
.pybloqs_life_foreground {{
  position: relative;
  z-index: 1;
  display: flex;
  justify-content: center;
  align-items: center;
  height: 100%;
  background: #ffffffa0;
  font-weight: bold;
  font-size: 2rem;
}}
"""

# This magic string is a compiled WASM program that exports a buffer called
# `BUFFER` of 50 by 50 pixels and a function called `go` which iterates
# Conway's game of life on the buffer.
WASM_B64 = """
AGFzbQEAAAABCQJgAABgAn9/AAMDAgABBQMBABEGIQR/AUGAgMAAC38AQYCAwAALfwBBlM7AAAt/
AEGgzsAACwczBQZtZW1vcnkCAAJnbwAABkJVRkZFUgMBCl9fZGF0YV9lbmQDAgtfX2hlYXBfYmFz
ZQMDCsoEAv4DAQl/I4CAgIAAQZDOAGsiACSAgICAAEEAKAKQzsCAAEEBaiEBQQAhAgNAIAJBgIDA
gABqIgMgAygCAEH///93SxCBgICAAAJAQQAoApDOwIAADQAgA0HPgvZ9Qc+C9gUgAUEBcRs2AgAg
AUGIoANxaUEBcSABQQF0ciEBCyACQQRqIgJBkM4ARw0AC0EAIQJBAEEAKAKQzsCAAEEBaiIDNgKQ
zsCAAAJAIANBCnANAANAIAAgAmogAkGAgMCAAGooAgBB////d0s2AgAgAkEEaiICQZDOAEcNAAtB
iLV/IQMgACgClAMhAiAAKAKQAyEEIAAoAgQhBSAAKALMASEGA0AgA0HEzMCAAGogBCACIgdqIAVq
IAZqIAAgA2oiAkH4ygBqKAIAaiACQYDLAGooAgAiBWogAkHAzABqKAIAaiACQcjMAGooAgAiCGog
AkGQzgBqKAIAIgJqIgRBfWpBAkkgBEEDRiAGGxCBgICAACAHIQQgCCEGIANBBGoiAw0AC0EAIQNB
gIDAgAAhAgNAIANBgIDAgABqIAFBAXFFEIGAgIAAIAIgAUEEcUUQgYCAgAAgAkHIzABqIAFBCHFF
EIGAgIAAIAFBiKADcWlBAXEgAUEBdHIhASACQQRqIQIgA0HIAWoiA0GQzgBHDQALCyAAQZDOAGok
gICAgAALSAEBfyAAKAIAIQICQAJAIAENACACQQJ2QYCAgP4DcUEDbEGAgIB4cSACQf///wdxciEC
DAELIAJBgICAeHIhAgsgACACNgIACwBOBG5hbWUACglsaWZlLndhc20BJwIAAmdvASBfWk40bGlm
ZTNzZXQxN2g2NzA1MDc4ZjY2OTZjMzg4RQcSAQAPX19zdGFja19wb2ludGVyAEUJcHJvZHVjZXJz
AQxwcm9jZXNzZWQtYnkBBXJ1c3RjJTEuODQuMC1uaWdodGx5ICg5MzIyZDE4M2YgMjAyNC0xMC0x
NCkASQ90YXJnZXRfZmVhdHVyZXMEKwptdWx0aXZhbHVlKw9tdXRhYmxlLWdsb2JhbHMrD3JlZmVy
ZW5jZS10eXBlcysIc2lnbi1leHQ=
""".replace("\n", "")

JS_STRING = f"""
async function life_init(canvas_id) {{
  const wasm_strbuffer = atob("{WASM_B64}");
  var wasm_codearray = new Uint8Array(wasm_strbuffer.length);
  for (var i in wasm_strbuffer) wasm_codearray[i] = wasm_strbuffer.charCodeAt(i);
  const {{ instance }} = await WebAssembly.instantiate(wasm_codearray);

  const width = 50;
  const height = 50;

  const canvas = document.getElementById(canvas_id);
  canvas.width = width;
  canvas.height = height;

  const buffer_address = instance.exports.BUFFER.value;
  const image = new ImageData(
      new Uint8ClampedArray(
          instance.exports.memory.buffer,
          buffer_address,
          4 * width * height,
      ),
      width,
  );

  const ctx = canvas.getContext("2d");

  const render = () => {{
      instance.exports.go();
      ctx.putImageData(image, 0, 0);
      setTimeout(render, 50);
  }};

  render();
}}
"""


class LifeLoadingBlock(pybloqs.BaseBlock):
    resource_deps = (Css(css_string=CSS_STRING, name="life_css"), JScript(script_string=JS_STRING, name="life_js"))

    def _write_contents(
        self, container, actual_cfg, id_gen: Iterator[str], resource_deps=None, static_output: Optional[bool] = None
    ) -> None:
        life_container = append_to(container, "div", **{"class": "pybloqs_life"})
        life_background = append_to(life_container, "div", **{"class": "pybloqs_life_background"})
        canvas = append_to(life_background, "canvas", id=next(id_gen))
        append_to(life_background, "div", **{"class": "grid"})
        foreground = append_to(life_container, "div", **{"class": "pybloqs_life_foreground"})
        foreground.string = "Loading"
        script = append_to(life_container, "script")
        script.string = f"setTimeout(() => life_init(\"{canvas['id']}\"),50);"



"""
Module for blocks with text-only content
"""

import textwrap

import bs4
import markdown

from pybloqs import BaseBlock
from pybloqs.html import parse


class Raw(BaseBlock):
    def __init__(self, contents: str, dedent: bool = True, **kwargs) -> None:
        """
        Writes out the content as raw text or HTML.

        :param contents: Raw text. Can contain arbitrary HTML.
        :param kwargs: Optional styling arguments. The `style` keyword argument has special
                       meaning in that it allows styling to be grouped as one argument.
                       It is also useful in case a styling parameter name clashes with a standard
                       block parameter.
        """
        super().__init__(**kwargs)

        if not isinstance(contents, str):
            raise ValueError("Expected string content type but got %s", type(contents))

        if dedent:
            contents = textwrap.dedent(contents)

        self._contents = self._process_raw_contents(contents)

    def _process_raw_contents(self, contents: str) -> str:
        return contents

    def _write_contents(self, container: bs4.Tag, *_args, **_kwargs) -> None:
        for child in list(parse(self._contents).children):
            container.append(child)


class Pre(Raw):
    """
    Renders the content with a fixed-width font, preserving whitespace.
    """

    container_tag = "pre"


class Span(Raw):
    """
    Renders a piece of text with formatting.
    """

    container_tag = "span"


class Markdown(Raw):
    """
    Renders Markdown into HTML content.
    """

    encoding = "UTF-8"

    def _process_raw_contents(self, contents: str) -> str:
        return markdown.markdown(contents, output_format="html")



import os
from io import StringIO
from typing import Any, Iterator, Optional

import bs4

from pybloqs.html import css_elem, js_elem
from pybloqs.util import encode_string, get_resource_path


class Resource:
    def __init__(
        self,
        file_name: Optional[str] = None,
        extension: str = "",
        content_string: Optional[str] = None,
        name: Optional[str] = None,
    ) -> None:
        """
        An external script dependency definition.

        :param file_name: Name of resource file included in static directory.
        :param extension: File extension of resource. Used to complement file_name if necessary.
        :param content_string: String with code or style, etc. provided as unicode string.
        :param name: Unique label used to identify duplicates. Only required with script_string.
        """
        # Do XOR for checking if either file_name or content_string and name are set
        if (file_name is None) == ((content_string is None) or (name is None)):
            raise ValueError("Please specify either resource file_name or content_string with name.")
        if file_name is None:
            self.name = name
            self.content_string = content_string
        else:
            self.name = os.path.splitext(file_name)[0]
            with open(self._local_path(file_name, extension), encoding="utf-8") as f:
                self.content_string = f.read()

    @classmethod
    def _local_path(cls, file_name: str, extension: str) -> str:
        """Generate the local path to the script using pkg_resources."""
        return get_resource_path(
            __package__, file_name if file_name.endswith(extension) else file_name + "." + extension
        )

    def write(self, parent: Optional[bs4.Tag], permit_compression: bool) -> bs4.Tag:
        raise NotImplementedError("write")

    def __hash__(self) -> int:
        return hash(self.name)

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, Resource):
            return False
        return self.name == other.name


class JScript(Resource):
    """
    Definition for external JS script dependencies.
    """

    # Global flag to turn off script encoding/compression everywhere
    global_encode = True

    def __init__(
        self,
        file_name: Optional[str] = None,
        script_string: Optional[str] = None,
        name: Optional[str] = None,
        encode: bool = True,
    ) -> None:
        """
        A JavaScript dependency definition. Ensures that multiple inclusions of the same function are handled safely.

        :param file_name: Name of resource file included in static directory.
        :param script_string: JS code provided as unicode string.
        :param name: Unique label used to identify duplicates. Only required with script_string.
        :param encode: Whether to compress and base64 encode the script.
        """
        super().__init__(file_name, "js", script_string, name)
        self.encode = encode

    def write(self, parent: Optional[bs4.Tag] = None, permit_compression: bool = True) -> bs4.Tag:
        stream = StringIO()

        # Wrapper to make accidental multiple inclusion if the same code (e.g. with different file names) safe to load.
        sentinel_var_name = "_pybloqs_load_sentinel_{}".format(self.name.replace("-", "_"))
        stream.write(f"if(typeof({sentinel_var_name}) == 'undefined'){{\n")

        if self.encode and permit_compression:
            self.write_compressed(stream, self.content_string)
        else:
            stream.write(self.content_string)

        # Second part of wrapper
        stream.write(f"\n{sentinel_var_name} = true;")
        stream.write("}")

        return js_elem(parent, stream.getvalue())

    @classmethod
    def write_compressed(cls, stream: StringIO, data: str) -> None:
        if cls.global_encode:
            stream.write('(async function(){var code = await RawDeflate.inflate("')
            stream.write(encode_string(data).decode())
            stream.write('"); blocksEval(code);})()')
        else:
            stream.write(data)


class Css(Resource):
    def __init__(
        self, file_name: Optional[str] = None, css_string: Optional[str] = None, name: Optional[str] = None
    ) -> None:
        """
        A CSS dependency definition. Will prevent adding resource with same name/file_name twice.

        :param file_name: Name of resource file included in static directory.
        :param css_string: CSS provided as unicode string.
        :param name: Unique label used to identify duplicates. Only required with script_string.
        """
        super().__init__(file_name, "css", css_string, name)

    def write(self, parent: Optional[bs4.Tag] = None, permit_compression: bool = True) -> bs4.Tag:
        return css_elem(parent, self.content_string)


class DependencyTracker:
    def __init__(self, *args) -> None:
        self._deps = list(args)

    def add(self, *resources: Resource) -> None:
        self._deps += [r for r in resources if r not in self._deps]

    def __iter__(self) -> Iterator[Resource]:
        return iter(self._deps)


# JS deflation script and the reporting core functionality is always registered
script_block_core = JScript("block-core", encode=False)
script_inflate = JScript("jsinflate", encode=False)

_registered_resources = DependencyTracker(script_block_core, script_inflate)


def register_interactive(*scripts: Resource) -> None:
    _registered_resources.add(*scripts)



from pybloqs.block.base import BaseBlock
from pybloqs.block.convenience import Block


class Box(BaseBlock):
    def __init__(self, contents, **kwargs) -> None:
        """
        Wrap the supplied content (can be anything that is supported by the basic blocks)
        in a container.

        :param contents: Content to wrap
        :param kwargs: Optional styling arguments. The `style` keyword argument has special
                       meaning in that it allows styling to be grouped as one argument.
                       It is also useful in case a styling parameter name clashes with a standard
                       block parameter.
        """
        super().__init__(**kwargs)

        # Blockify the content
        self._contents = Block(contents)

    def _write_contents(self, *args, **kwargs) -> None:
        self._contents._write_block(*args, **kwargs)


class Paragraph(Box):
    """
    Wraps the content in a paragraph.
    """

    container_tag = "p"



"""
Common email related functions used by various reports.
"""

import base64
import getpass
import os
import smtplib
import tempfile
import xml
from email import encoders
from email.message import Message
from email.mime.base import MIMEBase
from email.mime.image import MIMEImage
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from typing import Iterable, List, Optional, Tuple, Union

import html5lib
from html5lib import treebuilders

from pybloqs.config import user_config


def send(message: Message, recipients: Union[List[str], Tuple[str, ...]]) -> None:
    """
    Send an email to a list of recipients.

    - If the message is missing the From address it's set to the current user [#x]_.
    - If the message is missing the To field it's set to recipients [#x]_.
    - If add_footer is True and message is a subclass of MIMEBase the standard
      report email footer will be added before the email's sent.

    :param message: email.message.Message or subclass, eg email.mime.MIMEMultpart
    :param recipients: list of email addresses

    .. [#x] the original message is modified in these cases.
    """
    assert isinstance(message, Message)
    assert isinstance(recipients, (list, tuple))

    if not message["From"]:
        message["From"] = getpass.getuser()

    if not message["To"]:
        message["To"] = ",".join(recipients)

    s = smtplib.SMTP(**user_config["smtp_kwargs"])
    if "smtp_pre_login_calls" in user_config:
        smtp_pre_login_calls = user_config["smtp_pre_login_calls"]
        for method, kwargs in smtp_pre_login_calls:
            getattr(s, method)(**kwargs)
    if "smtp_login" in user_config:
        s.login(**user_config["smtp_login"])
    s.sendmail(message["From"], recipients, message.as_string())
    s.quit()


def _set_email_mime_types(
    dom: "xml.dom.minidom.Document", message: Optional[MIMEMultipart] = None, convert_to_ascii: bool = False
) -> MIMEMultipart:
    """
    Takes HTML dom object and constructs email object (MIMEMultipart)
    images referenced in html <img src=c:/test.png /> are attached as MIMEImage and
    reference in html part of email are replaced to reference these attachment
    in format <img src=cid:test.png />

    Returns - MIMEMultipart object
    :param dom: dom of the html message
    :param message: MIMEMultipart object (new instance created if None)
    :param convert_to_ascii: bool convert html format to ascii
    """
    # Create MIME
    if not message:
        message = MIMEMultipart()

    idx = 0
    subtype = None

    img_mimes = []
    # replace image paths with embedded images
    for img_tag in dom.getElementsByTagName("img"):
        src = img_tag.getAttribute("src")
        # Embedded images need special treatment
        if src.startswith("data:image"):
            name = f"{idx}_embedded"

            hdr = "data:image/"
            hdr_index = src.index(hdr) + len(hdr)
            imgdef = src[hdr_index:]
            subtype = imgdef[: imgdef.index(";")]

            # At the moment outlook doesn't support SVG so remove such images.
            # TODO: Rasterize the SVG on the fly for embedding.
            if subtype.lower() == "svg+xml":
                img_tag.parentNode.removeChild(img_tag)
                continue
            img_data = base64.b64decode(imgdef[(imgdef.index("base64,") + 7) :])
        elif src.startswith("http"):
            # Web hosted images aren't on the filesystem
            # and don't need to be changed
            continue
        else:
            # get a unique name for the image
            name = f"{idx}_{os.path.basename(src)}"

            with open(src, "rb") as fp:
                img_data = fp.read()

        img_mime = MIMEImage(img_data, _subtype=subtype)

        img_mime.add_header("Content-Disposition", "attachment", filename=name)
        img_mimes.append(img_mime)

        # replace the path with the embedded image
        img_tag.setAttribute("src", f"cid:{name}")

        idx += 1

    html = dom.toxml()
    if isinstance(html, str) and convert_to_ascii:
        html = html.encode("us-ascii", "ignore")
        message.attach(MIMEText(html, "html", "us-ascii"))
    else:
        message.attach(MIMEText(html, "html", "utf-8"))

    for img_mime in img_mimes:
        message.attach(img_mime)

    return message


def send_html_report(
    html_str: str,
    to: Iterable[str],
    subject: Optional[str] = None,
    attachments: Optional[Iterable[Union[str, Tuple[str, str]]]] = None,
    From: Optional[str] = None,
    Cc: Optional[Iterable[str]] = None,
    Bcc: Optional[Iterable[str]] = None,
    convert_to_ascii: bool = True,
) -> MIMEMultipart:
    """
    Email html report and embed any images.
    Extract html title to email subject.

    :return: MIMEMultipart object
    :param html_str: html string to send in the email body
    :param to: list of recipients
    :param subject: email subject
    :param attachments: a list of tuple(extension, filename)
    :param From: from, sender of the message
    :param Cc: cc recipient
    :param Bcc: bcc recipient
    :param convert_to_ascii: bool convert html format to ascii
    """
    # create the dom for querying/modifying the html document
    parser = html5lib.HTMLParser(tree=treebuilders.getTreeBuilder("dom"))
    dom = parser.parse(html_str)

    if not subject:
        # try to find title string in html
        for title in dom.getElementsByTagName("title"):
            if title.childNodes:
                subject = title.childNodes[0].data
                break

    # create the message and embed any images
    mime_msg = _set_email_mime_types(dom, convert_to_ascii=convert_to_ascii)

    if subject:
        mime_msg["Subject"] = subject

    if From:
        mime_msg["From"] = From

    mime_msg["To"] = ",".join(to)

    if Cc:
        mime_msg["Cc"] = ",".join(Cc)

    # set recipients to be union of to, cc'd and bcc'd addresses
    to = sorted(set(to) | set(Cc or []) | set(Bcc or []))

    if attachments is not None:
        for attachment_spec in attachments:
            try:
                filename_no_ext = subject
                fmt, content = attachment_spec
            except ValueError:
                filename_no_ext, fmt = os.path.splitext(os.path.split(attachment_spec)[-1])
                # Exclude the dot from the extension, gosh darn it!
                fmt = fmt[1:]
                if fmt == "":
                    raise ValueError("Attachment file name has no extension:", attachment_spec)

                with open(attachment_spec, "rb") as f:
                    content = f.read()

            attachment = MIMEBase("application", fmt)
            attachment.set_payload(content)

            encoders.encode_base64(attachment)

            filename_base = filename_no_ext or os.path.basename(tempfile.mktemp())
            attachment_spec = filename_base + "." + fmt

            attachment.add_header("Content-Disposition", "attachment", filename=attachment_spec)
            mime_msg.attach(attachment)

    send(mime_msg, to)
    return mime_msg



from io import StringIO
from typing import List, Tuple, Union

import pandas as pd

from pybloqs.plot import Chart, ColorAxis, Expr, Legend, Options3d, Plot, Scatter, Title, Tooltip, XAxis, YAxis
from pybloqs.plot import Heatmap as HeatmapPlot


class Heatmap(Plot):
    def __init__(self, data: Union[List, Tuple, pd.Series, pd.DataFrame], *args, **kwargs) -> None:
        super().__init__(data, HeatmapPlot(), flatten=True, *args, **kwargs)

        self._chart_cfg = self._chart_cfg.inherit_many(
            YAxis(title=None, reversed=True), XAxis(opposite=True), ColorAxis(min_color="#ffffff", max_color="#3060cf")
        )


class Corr(Heatmap):
    def __init__(self, data: Union[List, Tuple, pd.Series, pd.DataFrame], *args, **kwargs) -> None:
        super().__init__(data, *args, **kwargs)

        self._chart_cfg = self._chart_cfg.inherit_many(
            ColorAxis(min=0, max=1),
            Tooltip(
                formatter=Expr("""\
                function() {
                return '<b>' + this.series.xAxis.categories[this.point.x] + ' - '
                       + this.series.yAxis.categories[this.point.y] + '</b>: '
                       + ((this.point.value * 100) | 0) + '%';
                }
                """)
            ),
        )


class Surface(Plot):
    def __init__(self, data, *args, **kwargs) -> None:
        options = Options3d(
            enabled=True,
            alpha=kwargs.pop("alpha", 20),
            beta=kwargs.pop("beta", 0),
            depth=kwargs.pop("depth", 150),
            view_distance=kwargs.pop("view_distance", 10),
        )

        super().__init__(data, Scatter(), flatten=True, switch_zy=True, *args, **kwargs)

        self._chart_cfg = self._chart_cfg.inherit_many(
            Chart(options), Legend(enabled=False), YAxis(Title(text=None))
        ).override_many(Chart(zoom_type=None))

    def _write_plot_postprocess(self, chart_buf: StringIO) -> None:
        chart_buf.write("pybloqsSurfaceRotate(chart);")



import base64
import struct
import tempfile
from contextlib import contextmanager
from io import BytesIO, StringIO
from typing import Any, BinaryIO, Dict, Generator, Optional, Tuple, Union

import matplotlib.pyplot as plt
from matplotlib.artist import Artist
from matplotlib.figure import Figure

from pybloqs.block.base import BaseBlock
from pybloqs.block.convenience import add_block_types
from pybloqs.html import append_to, parse
from pybloqs.static import Css, JScript
from pybloqs.util import cfg_to_css_string

try:
    import plotly.offline as po
    from plotly.graph_objs import Figure as PlotlyFigure

    _PLOTLY_AVAILABLE = True
except ImportError:
    _PLOTLY_AVAILABLE = False

try:
    try:
        # Bokeh >= 3.1
        from bokeh.resources import Resources

        JSResources = Resources
        CSSResources = Resources
    except ImportError:
        try:
            from bokeh.resources import CSSResources, JSResources
        except ImportError:
            from bokeh.templates import CSSResources, JSResources
    try:
        from bokeh.plotting.figure import Figure as BokehFigure
    except ImportError:
        from bokeh.plotting import figure as BokehFigure
    from bokeh.embed.standalone import components
    from bokeh.io import export_png

    _BOKEH_AVAILABLE = True
except ImportError:
    _BOKEH_AVAILABLE = False


_MIME_TYPES = {"png": "png", "svg": "svg+xml"}

_PLOT_FORMAT: str = "png"
_PLOT_MIME_TYPE = _MIME_TYPES[_PLOT_FORMAT]
_PLOT_DPI: int = 100


@contextmanager
def plot_format(plot_format: Optional[str] = None, dpi: Optional[int] = None) -> Generator[None, None, None]:
    """
    Temporarily set the plot formatting settings

    :param plot_format: The plot format (e.g 'png')
    :type plot_format:  str
    :param dpi:         The DPI of the plots
    :type dpi:          int
    """
    old = get_plot_format()
    set_plot_format(plot_format, dpi)
    yield
    set_plot_format(*old)


def get_plot_format() -> Tuple[str, int]:
    """
    Get the current plot format parameters

    :return: tuple of format and dpi
    """
    return _PLOT_FORMAT, _PLOT_DPI


def set_plot_format(plot_format: Optional[str] = None, plot_dpi: Optional[int] = None) -> None:
    """
    Overwrite the current plot format settings

    :param plot_format: The plot format (e.g. 'png')
    :type  plot_format: str
    :param plot_dpi:    The DPI of the plots
    :type  plot_dpi:    int
    """
    global _PLOT_FORMAT
    global _PLOT_MIME_TYPE
    global _PLOT_DPI
    if plot_format is not None:
        _PLOT_FORMAT = plot_format
        _PLOT_MIME_TYPE = _MIME_TYPES[plot_format]

    if plot_dpi is not None:
        _PLOT_DPI = plot_dpi


class ImgBlock(BaseBlock):
    def __init__(
        self,
        data: Union[BaseBlock, bytes],
        mime_type: str = "png",
        width: Optional[str] = None,
        height: Optional[str] = None,
        img_style: Optional[Dict] = None,
        **kwargs,
    ) -> None:
        """
        Create a block containing an image. The dimensions can be sniffed from GIF
        and PNG data, for other formats, the `width` and `height` parameters can be
        used to specify the size.

        :param data: Either a block to convert to an image, or raw image data.
        :mime_type: The mime type of the image (e.g. "png" or "svg+xml")
        :param width: Image width override as string, e.g. "50px". If unspecified, the actual
                      image width is used.
        :param width: Image height override as string, e.g. "50px". If unspecified, the actual
                      image height is used.
        :param kwargs: Optional styling arguments. The `style` keyword argument has special
                       meaning in that it allows styling to be grouped as one argument.
                       It is also useful in case a styling parameter name clashes with a standard
                       block parameter.
        """
        # Wrapping a block in an image will render it
        if hasattr(data, "_write_block"):
            img_file = data.save(fmt=mime_type)
            with open(img_file, "rb") as f:
                data = f.read()

        self._img_data = base64.b64encode(data)
        self._mime_type = mime_type

        if width is None and height is None:
            if mime_type.lower() == "png":
                if struct.unpack("ccc", data[1:4]) != (b"P", b"N", b"G"):
                    raise ValueError("Image type is not png and does not match mime type")
                x, y = struct.unpack(">ii", data[16:24])
            elif mime_type.lower() == "gif":
                x, y = struct.unpack("<HH", data[6:10])
            else:
                raise ValueError(f"Can't determine image dimensions for mime type {mime_type}")

            width, height = (f"{x}px", f"{y}px")

        if img_style is None:
            img_styles = {}
        else:
            img_styles = img_style

        if width is not None:
            img_styles["width"] = width
        if height is not None:
            img_styles["height"] = height

        self._img_styles = img_styles

        super().__init__(**kwargs)

    def _write_contents(self, container, *args, **kwargs) -> None:
        src = StringIO()
        mime = f"data:image/{self._mime_type};base64,"
        src.write(mime)
        src.write(self._img_data.decode())

        img = append_to(container, "img", src=src.getvalue())

        if len(self._img_styles) > 0:
            img["style"] = cfg_to_css_string(self._img_styles)

    @staticmethod
    def from_file(img_file: Union[str, BinaryIO], **kwargs) -> "ImgBlock":
        """
        Load an image block from a file.

        :param img_file: File path or file-like object.
        :param kwargs: Arguments to pass to the `ImgBlock` constructor.
        :return: ImgBlock instance.
        """
        close_file = False

        if isinstance(img_file, str):
            img_file = open(img_file, "rb")  # noqa: SIM115
            close_file = True

        try:
            return ImgBlock(img_file.read(), **kwargs)
        finally:
            # Close the file in case it was opened in this function
            if close_file:
                img_file.close()


class PlotBlock(ImgBlock):
    def __init__(
        self,
        plot,
        close_plot: bool = True,
        bbox_inches="tight",
        width: Optional[str] = None,
        height: Optional[str] = None,
        **kwargs,
    ) -> None:
        """
        Create a block containing a matplotlib figure

        :param plot: A matplotlib figure, axes or artist object.
        :close_plot: Optional (default=True). Set to True to close the plot after it is
                     captured into an image and avoid lingering plot windows.
        :bbox_inches: Optional bounding box parameter for 'figure.savefig'.
        :param kwargs: Optional styling arguments. The `style` keyword argument has special
                       meaning in that it allows styling to be grouped as one argument.
                       It is also useful in case a styling parameter name clashes with a standard
                       block parameter.
        """
        if not isinstance(plot, Artist):
            raise ValueError("PlotBlock contents must be matplotlib Artist")

        if isinstance(plot, Figure):
            figure = plot
        elif isinstance(plot, Artist):
            figure = plot.get_figure()
        else:
            raise ValueError("Unexpected plot object type %s", type(plot))

        img_data = BytesIO()

        legends = []

        for ax in figure.get_axes():
            legend = ax.get_legend()
            if legend is not None:
                # Patch Legend get_window_extent since there seems to be a bug where
                # it is passed an unexpected renderer instance.
                _orig_get_window_extent = legend.get_window_extent

                def _patched_get_window_extent(*_) -> Any:
                    return _orig_get_window_extent()

                legend.get_window_extent = _patched_get_window_extent
                legends.append(legend)

        if len(figure.axes) == 0:
            # empty plot, disable bbox_inches to that savefig still works
            bbox_inches = None

        figure.savefig(
            img_data, dpi=_PLOT_DPI, format=_PLOT_FORMAT, bbox_extra_artists=legends, bbox_inches=bbox_inches
        )

        plt_width, plt_height = figure.get_size_inches()

        width = width or f"{plt_width:0.3f}in"
        height = height or f"{plt_height:0.3f}in"

        if close_plot:
            plt.close(figure)

        super().__init__(img_data.getvalue(), _PLOT_MIME_TYPE, width=width, height=height, **kwargs)

    def _to_static(self) -> BaseBlock:
        # Convert to a basic image block in case we contain 'dynamic' svg content
        return ImgBlock(self) if self._mime_type == "svg" else super()._to_static()


class PlotlyPlotBlock(BaseBlock):
    def __init__(self, contents, plotly_kwargs: Optional[Dict[str, Any]] = None, static_kwargs=None, **kwargs) -> None:
        """
        Writes out the content as raw text or HTML.

        :param contents: Plotly graphics object figure.
        :param plotly_kwargs: Kwargs that are passed to plotly plot function.
        :param static_kwargs: Kwargs that are passed to plotly function write_image() for static output.
        :param kwargs: Optional styling arguments. The `style` keyword argument has special
                       meaning in that it allows styling to be grouped as one argument.
                       It is also useful in case a styling parameter name clashes with a standard
                       block parameter.
        """
        self.resource_deps = [JScript(script_string=po.offline.get_plotlyjs(), name="plotly")]

        super().__init__(**kwargs)

        if not isinstance(contents, PlotlyFigure):
            raise ValueError("Expected plotly.graph_objs.graph_objs.Figure type but got %s", type(contents))

        plotly_kwargs = plotly_kwargs or {}
        self.static_kwargs = static_kwargs or {}
        prefix = "<script>if (typeof require !== 'undefined') {var Plotly=require('plotly')}</script>"
        self._fig = contents
        self._contents = prefix + po.plot(contents, include_plotlyjs=False, output_type="div", **plotly_kwargs)

    def _write_contents(self, container, *args, **kwargs) -> None:
        container.append(parse(self._contents))

    def _repr_html_(self) -> str:
        return self.render_html()

    def _to_static(self) -> ImgBlock:
        # Create a static png image for use e.g. in an email body
        with tempfile.NamedTemporaryFile(suffix=".png") as f:
            self._fig.write_image(f.name, **self.static_kwargs)
            static_block = ImgBlock.from_file(f)
        return static_block


class BokehPlotBlock(BaseBlock):
    def __init__(self, contents, static_kwargs=None, **kwargs) -> None:
        """
        Writes out the content as raw text or HTML.

        :param contents: Bokeh plotting figure.
        :param static_kwargs: Kwargs that are passed to Bokeh function export_png() for static output.
        :param kwargs: Optional styling arguments. The `style` keyword argument has special
                       meaning in that it allows styling to be grouped as one argument.
                       It is also useful in case a styling parameter name clashes with a standard
                       block parameter.
        """
        self.resource_deps = [JScript(script_string=s, name="bokeh_js") for s in JSResources().js_raw]
        self.resource_deps += [Css(css_string=s, name="bokeh_css") for s in CSSResources().css_raw]

        super().__init__(**kwargs)

        if not isinstance(contents, BokehFigure):
            raise ValueError("Expected bokeh.plotting.figure.Figure type but got %s", type(contents))

        self._fig = contents
        self.static_kwargs = static_kwargs or {}
        script, div = components(contents)
        self._contents = script + div

    def _write_contents(self, container, *args, **kwargs) -> None:
        container.append(parse(self._contents))

    def _to_static(self) -> ImgBlock:
        # Create a static png image for use e.g. in an email body
        with tempfile.NamedTemporaryFile(suffix=".png") as f:
            export_png(self._fig, filename=f.name, **self.static_kwargs)
            static_block = ImgBlock.from_file(f)
        return static_block


add_block_types(Artist, PlotBlock)
# If Plotly or Bokeh are not installed skip registration
if _PLOTLY_AVAILABLE:
    add_block_types(PlotlyFigure, PlotlyPlotBlock)

if _BOKEH_AVAILABLE:
    add_block_types(BokehFigure, BokehPlotBlock)



from typing import Collection, Mapping, Optional
from uuid import uuid4

from bs4.element import Tag

import pybloqs
from pybloqs.html import append_to
from pybloqs.server import BloqsProvider
from pybloqs.server.static import HTMX
from pybloqs.static import Css


class Tabs(pybloqs.BaseBlock):
    resource_deps = (
        HTMX,
        Css(
            css_string=""".blx_tab_bar{
                  border-bottom:1px solid #888;
                  padding-top:0.5em;
                  padding-bottom:0.5em:
                }
                .blx_tab_tab{
                  padding-right:0.5em;
                }
                .blx_tab_tab[aria-selected=true]{
                  font-weight: bold;
                }
                .tab_loading{
                  display:none;
                }
                .htmx-request.tab_loading{
                  display:block;
                }
                .htmx-request.tab_content{
                  display:none;
                }
                """,
            name="blx_tabs",
        ),
    )

    def __init__(
        self,
        contents: Optional[Mapping[str, BloqsProvider]] = None,
        options: Optional[Collection[str]] = None,
        provider: Optional[BloqsProvider] = None,
        loading_block: Optional[pybloqs.BaseBlock] = None,
        *args,
        **kwargs,
    ) -> None:
        from pybloqs.server.block.life_loading import LifeLoadingBlock

        super().__init__(*args, **kwargs)
        if contents is not None:
            if options is not None or provider is not None:
                raise ValueError("Cannot provide contents and either options or a provider")
            self.options = contents.keys()
            self.provider_generator = contents.__getitem__
        else:
            if options is None or provider is None:
                raise ValueError("Must provide both options and a provider")
            self.options = options
            self.provider_generator = provider.with_parameter
        self.loading_block = loading_block or LifeLoadingBlock()

    def _write_contents(self, container: Tag, actual_cfg, id_gen, resource_deps=None, static_output=None) -> None:
        tab_uid = f"tab{uuid4()}"

        tab_div = append_to(
            container,
            "div",
            **{
                "class": "blx_tab_bar",
                "hx-target": f"#{tab_uid}",
                "hx-on:htmx-before-request": """
                  event
                    .target
                    .parentNode
                    .querySelector('[aria-selected=true]')
                    .setAttribute('aria-selected', 'false');
                  event
                    .target
                    .setAttribute('aria-selected', 'true');
                """,
            },
        )

        content_div = append_to(
            container,
            "div",
            **{"id": tab_uid, "role": "tabpanel", "hx-trigger": "load", "class": "tab_content"},
        )
        loading_div = append_to(container, "div", **{"id": next(id_gen), "class": "tab_loading"})
        self.loading_block._write_block(
            loading_div, actual_cfg, id_gen=id_gen, resource_deps=resource_deps, static_output=static_output
        )

        for n, tab_title in enumerate(self.options):
            tab_contents = self.provider_generator(tab_title)
            tab = append_to(
                tab_div,
                "span",
                **{
                    "role": "tab",
                    "hx-get": tab_contents.url,
                    "hx-indicator": f"#{loading_div['id']}, #{content_div['id']}",
                    "class": "blx_tab_tab",
                    "aria-controls": tab_uid,
                },
            )
            tab.string = tab_title
            if n == 0:
                tab["aria-selected"] = "true"
                content_div["hx-get"] = tab_contents.url
                content_div["hx-indicator"] = f"#{loading_div['id']}, #{content_div['id']}"
            else:
                tab["aria-selected"] = "false"



from io import IOBase, StringIO
from typing import Sequence

import pybloqs.static as static
from pybloqs.plot.core import *  # noqa: F403
from pybloqs.plot.core import HIGHCHARTS_ALL, JScript


def add_highcharts_shim_to_stream(stream: IOBase, highcharts_all: Sequence[str]) -> None:
    """Generates JS to load Highcharts into Jupyter notebook. For human-readable sample output, see unit test."""
    highcharts_lower = [m.replace("-", "_") for m in highcharts_all]
    stream.write(
        """
        <script type='text/javascript'>
        if (typeof require !== 'undefined') {
        """
    )
    for lib_name, lib_name_lower in zip(highcharts_all[:-1], highcharts_lower[:-1]):
        stream.write(
            f"""
            require.undef("highcharts/{lib_name_lower}");
            define('highcharts/{lib_name_lower}', function(require, exports, module) {{
                {static.JScript(lib_name).content_string}
            }});
            """
        )
    stream.write(
        """
        require([{require_names}], function({param_names}) {{
            Highcharts = {highcharts};
        """.format(
            require_names=", ".join(f"'highcharts/{m}'" for m in highcharts_lower[:-1]),
            param_names=", ".join(highcharts_lower[:-1]),
            highcharts=highcharts_lower[0],
        )
    )
    stream.write("".join("{}(Highcharts);\n".format(m.replace("-", "_")) for m in highcharts_all[1:-1]))
    stream.write(static.JScript(highcharts_all[-1]).content_string)
    stream.write("\n")
    stream.writelines("\n".join(["window.Highcharts = Highcharts;", "});", "}", "</script>"]))


def interactive(verbose: bool = True) -> None:
    """Inject Highcharts JS into Jupyter notebook to use pybloqs.plot functions inside notebooks."""
    from IPython.core.display import display_html

    from pybloqs.html import id_generator_uuid, set_id_generator

    set_id_generator(id_generator_uuid)

    stream = StringIO()

    stream.write(str(JScript("block-core", encode=False).write()))
    stream.write(str(JScript("jsinflate", encode=False).write()))
    add_highcharts_shim_to_stream(stream, HIGHCHARTS_ALL)

    display_html(stream.getvalue(), raw=True)

    if verbose:
        stream.write("<div>Interactive mode initialized successfully</div>")



from datetime import datetime
from io import StringIO
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    Iterator,
    List,
    Literal,
    Optional,
    Sequence,
    TextIO,
    Tuple,
    Type,
    Union,
)

import bs4
import numpy as np
import pandas as pd
from pandas.core.generic import NDFrame

from pybloqs.block.base import BaseBlock
from pybloqs.block.image import ImgBlock
from pybloqs.html import append_to, js_elem
from pybloqs.static import JScript
from pybloqs.util import Cfg, camelcase, dt_epoch_msecs, np_dt_epoch_msec

# Sets of plots based on dimensionality
_univariate_plots = {"area", "areaspline", "column", "flags", "line", "scatter", "spline", "pie", "gauge", "funnel"}
_range_plots = {"arearange", "areasplinerange", "candlestick", "columnrange", "ohlc"}

HIGHCHARTS_MAIN = "highstock"
HIGHCHARTS_MODULES = ["highcharts-more", "highcharts-3d", "heatmap", "funnel", "exporting", "export-data"]
HIGHCHARTS_PYBLOQS = "highcharts-pybloqs"
# Ordering has to be: main, modules, pybloqs-specific
HIGHCHARTS_ALL = [HIGHCHARTS_MAIN, *HIGHCHARTS_MODULES, HIGHCHARTS_PYBLOQS]


class Expr:
    """
    Represents a javascript expression as a string.
    """

    def __init__(self, fun_str: str) -> None:
        self.fun_str = fun_str

    def write_jscript(self, stream: TextIO) -> None:
        stream.write(self.fun_str)

    def __repr__(self) -> str:
        return f"Expr('''{self.fun_str}''')"


class _PlotDim(Expr):
    """
    Object to hold javascript evaluated plot dimensions.

    TODO: Replace current eager evaluation with a simple lazy tree.
    """

    def __add__(self, other: Any) -> "_PlotDim":
        return self._construct_arith("+", other)

    def __sub__(self, other: Any) -> "_PlotDim":
        return self._construct_arith("-", other)

    def __mul__(self, other: Any) -> "_PlotDim":
        return self._construct_arith("*", other)

    def __div__(self, other: Any) -> "_PlotDim":
        return self._construct_arith("/", other)

    def __truediv__(self, other: Any) -> "_PlotDim":
        return self.__div__(other)

    def __mod__(self, other: Any) -> "_PlotDim":
        return self._construct_arith("%", other)

    def __pow__(self, power, modulo=None) -> "_PlotDim":
        if modulo is not None:
            raise NotImplementedError

        return self._construct_arith("^", power)

    def _construct_arith(self, op: str, other: Any) -> "_PlotDim":
        return _PlotDim(self.fun_str + op + str(other))


class _PlotOpts(Cfg):
    pass


class Plot(BaseBlock):
    """
    A composable chart. When the input data is a list of plot objects, they will be merged into a single
    figure.
    """

    resource_deps = tuple(JScript(m) for m in HIGHCHARTS_ALL)

    def __init__(self, data: Union[List, Tuple, pd.Series, pd.DataFrame], *args, **kwargs) -> None:
        """
        Create a chart or composite chart from the supplied data.

        :param data: List, tuple, pandas.Series/DataFrame to use as chart data.
                     In case `data` is a list of Plot objects, a composite chart will be constructed.
        :param chart_cls: The chart class to use. Available values are "Chart" or "StockChart".
                          StockCharts have extra features like a navigator pane and special handling
                          for timestamps. A sensible default will be chosen based on the data by default.
        :param flatten: When set to True, the data will be flattened and categories will be
                        extracted from labelled data automatically.
                        Useful when creating plots where both axes need category labels (e.g. heatmaps).
        :param args: Chart level configuration. Axes definitions will be applied to subplots
                     that do not have custom axes. Plots option sets supplied here will
                     be used by default for any subplots that do not specify their own.
        :param kwargs: Optional styling arguments. The `style` keyword argument has special
                       meaning in that it allows styling to be grouped as one argument.
                       It is also useful in case a styling parameter name clashes with a standard
                       block parameter.
        """
        chart_cls = kwargs.pop("chart_cls", self._choose_chart_class(data))
        flatten = kwargs.pop("flatten", False)
        switch_zy = kwargs.pop("switch_zy", False)

        super().__init__(**kwargs)

        chart_cfg, plot_cfg = self._parse_args(args)

        # In case we got a list/tuple of Plots, create a composite chart.
        if isinstance(data, (list, tuple)) and isinstance(data[0], (Plot, NDFrame)):
            # Parse out any axis defaults
            axis_defaults = {}
            for ax_name in ["x_axis", "y_axis"]:
                axis_cfg = chart_cfg.pop(ax_name, None)
                if axis_cfg is not None:
                    axis_defaults[ax_name] = axis_cfg

            axes = Cfg(x_axis=[], y_axis=[])

            chart_series = []

            for plot_data in data:
                # Convert data to a plot in case the it was not an actual plot instance
                # (but a pandas object for example).
                if not isinstance(plot_data, Plot):
                    plot_data = Plot(plot_data)
                    # If no chart class was defined yet, choose one based on the data.
                if chart_cls is None:
                    chart_cls = plot_data._chart_cls

                # Extract the chart configuration.
                subchart_cfg = plot_data._chart_cfg

                # Extract the series from the subchart config
                subchart_series = subchart_cfg.series

                # Check for axis definitions.
                def _set_axis(axis_name) -> None:
                    # Get a custom axis, or use/create a default one.
                    axis = subchart_cfg.setdefault(axis_name, None) or axis_defaults.setdefault(axis_name, Cfg())
                    if isinstance(axis, (list, tuple)):
                        if len(axis) != 1:
                            raise ValueError("There can be at most one sub-chart axis definition for each axis.")

                        axis = axis[0]

                    # Get the index of this axis from the list of known axes. Optionally, add a new entry.
                    axis_coll = axes[axis_name]
                    try:
                        axis_idx = axis_coll.index(axis)
                    except ValueError:
                        axis_coll.append(axis)
                        axis_idx = len(axis_coll) - 1

                    for series_cfg in subchart_series:
                        series_cfg[axis_name] = axis_idx

                _set_axis("x_axis")
                _set_axis("y_axis")

                # Add the subplot series to the figure series.
                chart_series.extend(subchart_series)

            chart_cfg = chart_cfg.override(axes)
            chart_cfg.series = chart_series
        else:
            if flatten:
                data, chart_cfg = self._flatten_data(data, chart_cfg, switch_zy=switch_zy)

            chart_cfg.series = self._construct_plot_series(data, plot_cfg)

        chart_cfg = self._set_chart_defaults(chart_cfg, chart_cls)

        self._chart_cfg = chart_cfg
        self._chart_cls = chart_cls

    def inherit(self, *args: Cfg) -> None:
        self._chart_cfg = self._chart_cfg.inherit_many(*args)

    @staticmethod
    def _flatten_data(
        data: Union[Sequence, NDFrame], chart_cfg: Cfg, switch_zy: bool = False
    ) -> Tuple[Union[Sequence, NDFrame], Cfg]:
        plot_axes_def = [(0, XAxis), (1, YAxis)]

        # Inject categories into the axis definitions of the plot
        if isinstance(data, NDFrame):
            for i, plot_axis in plot_axes_def[: data.ndim]:
                categories = data.axes[i]
                # Skip numeric indices
                if not categories.is_numeric():
                    chart_cfg = chart_cfg.inherit_many(plot_axis(categories=list(categories)))

        data = [[*index, value] for index, value in list(np.ndenumerate(data))]

        if switch_zy:
            for i in range(len(data)):
                tmp = data[i][-1]
                data[i][-1] = data[i][-2]
                data[i][-2] = tmp

        return data, chart_cfg

    @staticmethod
    def _set_chart_defaults(chart_cfg: Cfg, chart_cls: Literal["StockChart", "Chart"]) -> Cfg:
        if chart_cls == "StockChart":
            chart_cfg = chart_cfg.inherit_many(
                Chart(zoom_type="x"), PlotOptions(Series(States(Hover(enabled=False, halo=False))))
            )
        else:
            chart_cfg = chart_cfg.inherit_many(Chart(zoom_type="xy"))

        # Stockcharts forces the Y Axis to be on the right side by default and it
        # does not honor overrides set globally in JS. Fix it here.
        if "y_axis" not in chart_cfg:
            return chart_cfg.inherit(YAxis(opposite=False))
        else:
            base_cfg = Cfg(opposite=False)

            # In case we have a single y axis, just set the defaults on it
            if isinstance(chart_cfg.y_axis, Cfg):
                chart_cfg.y_axis = chart_cfg.y_axis.inherit(base_cfg)
            else:
                y_axes = chart_cfg.y_axis

                # Set the default on all y axes
                for i in range(len(y_axes)):
                    y_axes[i] = y_axes[i].inherit(base_cfg)

            return chart_cfg

    @staticmethod
    def _parse_args(
        args: Iterable[Union[Type, _PlotOpts, Any]], allowed_cfg_types: Tuple[Type, ...] = (Cfg,)
    ) -> Tuple[Cfg, Union[None, Cfg, _PlotOpts]]:
        """
        Parse the supplied argument list into a configuration object.

        :param args: Argument list.
        :param allowed_cfg_types: The listed configuration object types will be allowed.
        :return: Tuple of (chart configuration object, plot configuration object OR None)
        """
        plot_cfg = None
        configs = Cfg()
        for arg in args:
            # Construct an instance in case the config object is passed in as a type
            if isinstance(arg, type):
                arg = arg()

            if isinstance(arg, _PlotOpts):
                if plot_cfg:
                    plot_cfg = plot_cfg.inherit(arg)
                else:
                    plot_cfg = arg
            elif isinstance(arg, allowed_cfg_types):
                configs = configs.inherit(arg)
            else:
                raise ValueError(f"{arg} is not recognized as a plot or chart configuration object")
        return configs, plot_cfg

    @staticmethod
    def _construct_plot_series(data, plot_cfg: Optional[Cfg] = None) -> List[Cfg]:
        """
        Construct a list of series configurations.

        :param data: The data to bundle into chart series.
        :param plot_cfg: Optional. Existing plot configuration. If unspecified, a sensible
                         default will be chosen based on the data type.
        :return: List of chart series.
        """

        def _decompose_l1(cfg: Cfg) -> List[Cfg]:
            return [cfg.override_many(data=value).inherit_many(name=key) for key, value in data.items()]

        def _decompose_l2(cfg: Cfg) -> List[Cfg]:
            component_series = []

            for k1, v1 in data.items():
                for k2, v2 in v1.items():
                    component_series.append(cfg.override_many(data=v2).inherit_many(name=f"{k1} - {k2}"))

            return component_series

        def _wrap(cfg: Cfg, name: str) -> List[Cfg]:
            cfg = cfg.override(Cfg(data=data))
            if "name" not in cfg and name is not None:
                cfg.name = name
            return [cfg]

        data_labelled = True

        # Construct a series configuration, with optionally choosing a default plot configuration as well
        if isinstance(data, pd.Series):
            plot_cfg = plot_cfg or Line()
            series = _wrap(plot_cfg, data.name)
        elif isinstance(data, pd.DataFrame):
            plot_cfg = plot_cfg or Line()
            if plot_cfg.type in _univariate_plots:
                series = _decompose_l1(plot_cfg)
            else:
                series = _wrap(plot_cfg, None)
        else:
            plot_cfg = plot_cfg or Line()
            series = _wrap(plot_cfg, getattr(data, "name", None))
            data_labelled = False

        plot_cfg.check_array_shape(data, data_labelled)

        return series

    @staticmethod
    def _choose_chart_class(data: Union[pd.Series, NDFrame, Sequence]) -> Literal["StockChart", "Chart"]:
        """
        Tries to guess the appropriate chart class based on the data.
        """
        # In case the data has an index and the first entry is a datetime type, return a stock chart
        # specialized for viewing time series.
        if isinstance(data, pd.Series):
            if isinstance(data.index[0], (np.datetime64, datetime)):
                return "StockChart"
        elif isinstance(data, NDFrame):
            for labels in data.axes:
                if isinstance(labels[0], (np.datetime64, datetime)):
                    return "StockChart"
        elif hasattr(data, "__getitem__"):
            try:
                if isinstance(data[0], Plot):
                    return data[0]._chart_cls
                if (
                    (len(data[0]) > 1)
                    and isinstance(data[0], (list, tuple))
                    and isinstance(data[0][0], (np.datetime64, datetime))
                ):
                    return "StockChart"
            except TypeError:
                pass

        return "Chart"

    def _write_contents(
        self, container, actual_cfg, id_gen: Iterator[str], static_output: bool = False, **kwargs
    ) -> None:
        plot_container = append_to(container, "div")
        plot_container["id"] = plot_container_id = next(id_gen)

        # Write the config to the plot target as well
        self._write_container_attrs(plot_container, actual_cfg)

        self._write_plot(plot_container, plot_container_id, id_gen, static_output)

    def _write_plot(self, container: bs4.Tag, container_id: str, id_gen: Iterator[str], static_output: bool) -> None:
        """
        Write out the chart construction machinery.
        """
        js_timer_var_name = "_ins_timer_" + next(id_gen)

        # Plumbing to make sure script rendering waits until the container element is ready
        stream = StringIO()

        chart_cfg = self._chart_cfg.override(Chart(render_to=container_id))

        if static_output:
            # Chart load wait handles for static output.
            stream.write(f"registerWaitHandle('{container_id}');")
            overrides = [
                Chart(Events(load=Expr(f"function(){{setLoaded('{container_id}');}}"))),
                Exporting(enabled=False),
                Navigator(enabled=False),
                Scrollbar(enabled=False),
                PlotOptions(Series(enable_mouse_tracking=False, shadow=False, animation=False)),
                RangeSelector(enabled=False),
            ]

            chart_cfg = chart_cfg.override_many(*overrides)

        stream.write(
            f"var {js_timer_var_name}=setInterval(function(){{"
            f"var container=document.getElementById('{container_id}');"
            f"if(container){{clearInterval({js_timer_var_name});"
        )

        stream.write("var cfg=")
        self._write_dict(stream, chart_cfg)
        stream.write(";")

        stream.write("var chart = new Highcharts." + self._chart_cls + "(cfg);")

        self._write_plot_postprocess(stream)

        stream.write("}},10);")

        js_elem(container, stream.getvalue())

    def _write_plot_postprocess(self, chart_buf: StringIO) -> None:
        pass

    def _write_value(self, stream, value) -> None:
        """
        Write the supplied value to the stream.
        """
        # WARNING: The bool case must come before the (int, float) since python bools are ints as well.
        if isinstance(value, np.datetime64):
            stream.write(str(np_dt_epoch_msec(value)))
        elif isinstance(value, datetime):
            stream.write(str(dt_epoch_msecs(value)))
        elif isinstance(value, (bool, np.bool_)):
            stream.write("true" if value else "false")
        elif isinstance(value, (int, int, float, np.number)):
            if np.isnan(value) or np.isinf(value):
                stream.write("null")
            else:
                stream.write(str(value))
        elif isinstance(value, str):
            stream.write("'" + value + "'")
        elif isinstance(value, str):
            stream.write("'" + str(value) + "'")
        elif isinstance(value, dict):
            self._write_dict(stream, value)
        elif isinstance(value, (list, tuple, set)):
            self._write_iterable(stream, value)
        elif isinstance(value, pd.Series):
            self._write_iterable(stream, zip(value.index, value.values))
        elif isinstance(value, pd.DataFrame):
            labels = value.index
            values = value.values

            # Merge DFrame into a single list of lists
            merged = []
            for i in range(len(labels)):
                label = labels[i]
                ndval = values[i]

                if isinstance(label, tuple):
                    merged.append(tuple(*label, *ndval))
                else:
                    merged.append([label, *ndval])

            self._write_iterable(stream, merged)
        elif isinstance(value, (np.ndarray, pd.Index)):
            self._write_iterable(stream, value)
        elif hasattr(value, "write_jscript"):
            value.write_jscript(stream)
        elif value is None:
            stream.write("null")
        else:
            raise ValueError("Unhandled config item type " + str(type(value)))

    def _write_dict(self, stream, dct) -> None:
        """
        Write out a dictionary.
        """
        stream.write("{")
        for i, item in enumerate(dct.items()):
            # If this is not the first item at this level, prepend a comma
            if i > 0:
                stream.write(",")

            key, value = item
            # camelCase the key as appropriate
            stream.write(camelcase(key) + ":")
            self._write_value(stream, value)
        stream.write("}")

    def _write_iterable(self, stream, iterable) -> None:
        """
        Write out an iterable.
        """
        stream.write("[")
        for i, item in enumerate(iterable):
            # If this is not the first item at this level, prepend a comma
            if i > 0:
                stream.write(",")

            self._write_value(stream, item)

        stream.write("]")

    def _to_static(self) -> ImgBlock:
        return ImgBlock(self)


def _make_chart_cfg(name: str, *def_args, **def_kwargs) -> Callable:
    """
    Creates a chart configuration group. Uniqueness is ensured by attaching an UUID
    base __id keyword.
    """

    def _builder(*args: Dict, **kwargs) -> Cfg:
        for arg in args:
            kwargs.update(arg)

        kwargs["__id"] = hash(name)
        return Cfg({name: Cfg(kwargs).inherit_many(*def_args, **def_kwargs)})

    return _builder


# Main Chart configuration groups.
Chart = _make_chart_cfg("chart")
Colors = lambda colors: Cfg({"colors": colors})  # Colors is an array and not an option group
Credits = _make_chart_cfg("credits")
Exporting = _make_chart_cfg("exporting")
Labels = _make_chart_cfg("labels")
Legend = _make_chart_cfg("legend")
Loading = _make_chart_cfg("loading")
Pane = _make_chart_cfg("pane")
Navigation = _make_chart_cfg("navigation")
Navigator = _make_chart_cfg("navigator")
PlotOptions = _make_chart_cfg("plot_options")
RangeSelector = _make_chart_cfg("range_selector")
Scrollbar = _make_chart_cfg("scrollbar")
Subtitle = _make_chart_cfg("subtitle")
Title = _make_chart_cfg("title")
Tooltip = _make_chart_cfg("tooltip")
TooltipPct = _make_chart_cfg(
    "tooltip",
    value_decimals=3,
    point_format='<span style="color:{series.color}">{series.name}</span>:' " <b>{point.y}%</b><br/>",
)

XAxis = _make_chart_cfg("x_axis")
YAxis = _make_chart_cfg("y_axis")
YAxisPct = _make_chart_cfg("y_axis", Labels(formatter=Expr("function(){return (this.value>0?'+':'')+this.value+'%';}")))
ZAxis = _make_chart_cfg("z_axis")

# Secondary Chart configuration groups
ColorAxis = _make_chart_cfg("color_axis")
Options3d = _make_chart_cfg("options3d")
Events = _make_chart_cfg("events")
ResetZoomButton = _make_chart_cfg("reset_zoom_button")
Buttons = _make_chart_cfg("buttons")
ExportButton = _make_chart_cfg("export_button")
PrintButton = _make_chart_cfg("print_button")
DataClasses = lambda items: Cfg({"data_classes": items})  # DataClasses is an array and not an option group
Items = lambda items: Cfg({"items": items})  # Items is an array and not an option group
Frame = _make_chart_cfg("frame")
Back = _make_chart_cfg("back")
Bottom = _make_chart_cfg("bottom")
Side = _make_chart_cfg("side")
Style = _make_chart_cfg("style")

# Series Configuration groups
Series = _make_chart_cfg("series")
DataLabels = _make_chart_cfg("data_labels")
Marker = _make_chart_cfg("marker")
States = _make_chart_cfg("states")
Hover = _make_chart_cfg("hover")
Select = _make_chart_cfg("select")
Point = _make_chart_cfg("point")
Dial = _make_chart_cfg("dial")
Pivot = _make_chart_cfg("pivot")
DataGrouping = _make_chart_cfg("data_grouping")

# Axis Configuration groups
PlotBands = lambda items: Cfg({"plot_bands": items})  # PlotBands is an array and not an option group
PlotLines = lambda items: Cfg({"plot_lines": items})  # PlotLines is an array and not an option group
StackLabels = _make_chart_cfg("stack_labels")


def _make_plot_opts(plot_type: str, rank: int) -> Type[_PlotOpts]:
    class _SpecPlotOpts(_PlotOpts):
        def __init__(self, *args, **kwargs) -> None:
            super().__init__(*args, **kwargs)
            self.type = plot_type

        def check_array_shape(self, arr: Union[Sequence, NDFrame], is_labelled: bool) -> None:
            ndim, shape = _sniff_data_dim(arr)

            if ndim == 1:
                if rank > 1:
                    raise ValueError("Multivariate plots need at least two dimensional input.")

                # For single dimensional data, the minor axis length is 1
                minor_axis_length = 1
            else:
                minor_axis_length = shape[-1]

                # For non labelled data, we substract one from the minor axis length to account
                # for the first item implicitly becoming an label.
                if not is_labelled:
                    minor_axis_length = max(minor_axis_length - 1, 1)

            if minor_axis_length < rank:
                raise ValueError(
                    f"Supplied array length for plot type {self.type} must be "
                    f"{rank} on the minor axis (got {minor_axis_length})."
                )

    return _SpecPlotOpts


def _sniff_data_dim(data: Union[Sequence, NDFrame]) -> Tuple[int, List[int]]:
    try:
        return data.ndim, data.shape
    except AttributeError:
        # arr is not a numpy array.
        return _sniff_list_dim(data)


def _sniff_list_dim(data: Sequence) -> Tuple[int, List[int]]:
    dim = []

    def _sniff_rec(item: Sequence) -> None:
        # noinspection PyBroadException
        try:
            dim.append(len(item))
            _sniff_rec(item[0])
        except Exception:
            pass

    _sniff_rec(data)

    return len(dim), dim


# Plot options
# Univariate
Area = _make_plot_opts("area", 1)
Areaspline = _make_plot_opts("areaspline", 1)
Column = _make_plot_opts("column", 1)
Bar = _make_plot_opts("bar", 1)
Flags = _make_plot_opts("flags", 1)
Line = _make_plot_opts("line", 1)
Scatter = _make_plot_opts("scatter", 1)
Spline = _make_plot_opts("spline", 1)
Pie = _make_plot_opts("pie", 1)
Gauge = _make_plot_opts("gauge", 1)
Errorbar = _make_plot_opts("errorbar", 1)
Funnel = _make_plot_opts("funnel", 1)

# Bivariate
Arearange = _make_plot_opts("arearange", 2)
Areasplinerange = _make_plot_opts("areasplinerange", 2)
Bubble = _make_plot_opts("bubble", 2)
Heatmap = _make_plot_opts("heatmap", 2)
Columnrange = _make_plot_opts("columnrange", 2)

# Quadrivariate
Boxplot = _make_plot_opts("boxplot", 4)
Candlestick = _make_plot_opts("candlestick", 4)
Ohlc = _make_plot_opts("ohlc", 4)


class ChartPeriods:
    second = 1000
    minute = 60 * second
    hour = 60 * minute
    day = 24 * hour
    month = 30 * day
    year = 12 * month



import enum
from collections import OrderedDict
from typing import Callable, Dict, List, Optional, Tuple, Union, cast

import pandas as pd

import pybloqs.block.table_formatters as pbtf
from pybloqs.block import colors


class FormatterType(enum.Enum):
    page_break = enum.auto()
    table = enum.auto()
    header = enum.auto()
    index = enum.auto()
    decimal = enum.auto()
    bps = enum.auto()
    percent = enum.auto()
    int = enum.auto()
    dollar = enum.auto()
    date = enum.auto()
    replace_nans = enum.auto()
    truncate_contents_with_ellipsis = enum.auto()
    heat_map = enum.auto()
    threshold = enum.auto()
    divider_lines = enum.auto()
    hide_cells = enum.auto()
    hide_insignificant = enum.auto()
    color_background = enum.auto()
    stripe_background = enum.auto()
    total = enum.auto()


class FormatterBuilder:
    def __init__(self) -> None:
        self._formatters: Dict[FormatterType, List[pbtf.TableFormatter]] = OrderedDict()

    def add_formatter(self, name: FormatterType, formatter: pbtf.TableFormatter) -> "FormatterBuilder":
        if name not in self._formatters:
            self._formatters[name] = [formatter]
        else:
            self._formatters[name].append(formatter)
        return self

    def remove_formatter(self, name: FormatterType) -> "FormatterBuilder":
        if name in self._formatters:
            del self._formatters[name]
        return self

    def replace_formatter(self, name: FormatterType, formatter: pbtf.TableFormatter) -> None:
        self.remove_formatter(name)
        self.add_formatter(name, formatter)

    @property
    def formatters(self) -> List[pbtf.TableFormatter]:
        return [fmt for fmts in self._formatters.values() for fmt in fmts if fmt is not None]


class CommonTableFormatterBuilder(FormatterBuilder):
    def __init__(
        self,
        data: pd.DataFrame,
        use_defaults=True,
        page_break: bool = True,
        table_width: Optional[str] = "98%",
        table_align: Optional[str] = "center",
        align_header: Optional[str] = "right",
        header_divider: bool = True,
        hide_index: bool = False,
        index_width: Optional[str] = "5%",
        align_index: Optional[str] = "left",
        index_bold: bool = True,
        column_width: Optional[str] = "5%",
        align_column: Optional[str] = "right",
        font_style: Optional[str] = "Calibri, 'Trebuchet MS', Helvetica, sans-serif;",
        font_size: Optional[int] = 6,
        cell_padding: Optional[int] = 5,
        replace_nans: Union[str, bool, None] = "",
        hide_insignificant: bool = True,
        truncate_index: bool = True,
    ) -> None:
        """
        Create a table formatter builder, optionally set with various default formatters.

        :param data
            dataframe that will be rendered into table - this must be provided
        :param use_defaults
            use preset formatters and default values to render table - defaults are used by default - if disabled
            then all following parameters are irrelevant...
        :param page_break
            whether to split table across pages, defaults to split
        :param table_width
            constrain width of table - e.g. '98%'; None to leave unconstrained
        :param table_align
            align table within page - e.g. 'center', 'right' or 'left'; None to leave unconstrained
        :param align_header
            align header values - e.g. 'center', 'right' or 'left'; None to leave unconstrained
        :param header_divider
            include a dividing line between header row and body of the table
        :param hide_index
            do not show index in rendered table
        :param index_width
            specify column width as pct of table - e.g. '15%'; None to leave unconstrained
        :param align_index
            align index values - e.g. 'center', 'right' or 'left'; None to leave unconstrained
        :param index_bold
            whether the index values should be rendered bold
        :param column_width
            specify column width as pct of table - e.g. '5%'; None to leave unconstrained
        :param align_column
            align table values - e.g. 'center', 'right' or 'left'; None to leave unconstrained
        :param font_style
            specify font style for index values - e.g. "Helvetica, sans-serif;" or None to leave unconstrained
        :param font_size
            specify font size for index values - e.g. 6; None to leave unconstrained
        :param cell_padding
            specify left and right padding for index values - e.g. 5; None to leave unconstrained
        :param replace_nans
            replace nan values with specified string - e.g. blank, pass False or None to disable
        :param hide_insignificant
            replace zero values with blank string
        :param truncate_index
            shorten index values to fit into index column, if done then there will be a trailing ellipses
        """
        super().__init__()
        if data is None:
            raise ValueError("You must provide the dataframe to be formatted!")
        self._data = data
        self._use_defaults = use_defaults
        self._page_break = page_break
        self._table_width = table_width
        self._table_align = table_align
        self._align_header = align_header
        self._header_divider = header_divider
        self._hide_index = hide_index
        self._index_width = index_width
        self._align_index = align_index
        self._index_bold = index_bold
        self._column_width = column_width
        self._align_column = align_column
        self._font_style = font_style
        self._font_size = font_size
        self._cell_padding = cell_padding

        if use_defaults:
            self.set_page_break(page_break)
            self.format_table(
                table_width=table_width,
                table_align=table_align,
                column_width=column_width,
                align_column=align_column,
                font_style=font_style,
                font_size=font_size,
                cell_padding=cell_padding,
            )
            self.format_header(
                align_header=align_header,
                header_divider=header_divider,
                font_style=font_style,
                font_size=font_size,
                cell_padding=cell_padding,
            )
            if self._hide_index:
                self.hide_index()
            else:
                self.format_index(
                    index_width=index_width,
                    align_index=align_index,
                    index_bold=index_bold,
                    font_style=font_style,
                    font_size=font_size,
                    cell_padding=cell_padding,
                    table_width=table_width,
                )
            if replace_nans is True:
                self.replace_nans()
            elif replace_nans is not False and replace_nans is not None:
                self.replace_nans(cast(str, replace_nans))
            if hide_insignificant:
                self.hide_insignificant()
            if truncate_index:
                self.truncate_contents_with_ellipsis(columns=[pbtf.INDEX_COL_NAME])

    def set_page_break(self, page_break: bool = True) -> "CommonTableFormatterBuilder":
        """
        Set page break behaviour to either split the table across pages and repeat the header or force
        the table onto a single page.

        :param page_break
            whether to split table across pages, defaults to split

        :returns: builder
        """
        if page_break:
            self.add_formatter(FormatterType.page_break, pbtf.FmtPageBreak(no_break=False, repeat_header=True))
        else:
            self.add_formatter(FormatterType.page_break, pbtf.FmtPageBreak(no_break=True, repeat_header=False))
        return self

    def format_table(
        self,
        table_width: Optional[str],
        table_align: Optional[str],
        column_width: Optional[str],
        align_column: Optional[str],
        font_style: Optional[str],
        font_size: Optional[int],
        cell_padding: Optional[int],
    ) -> "CommonTableFormatterBuilder":
        """
        Set formatters for content of table

        :param table_width
            constrain width of table - e.g. '98%'; None to leave unconstrained
        :param table_align
            align table within page - e.g. 'center', 'right' or 'left'; None to leave unconstrained
        :param column_width
            specify column width as pct of table - e.g. '5%'; None to leave unconstrained
        :param align_column
            align table values - e.g. 'center', 'right' or 'left'; None to leave unconstrained
        :param font_style
            specify font style for table values - e.g. rs.DEFAULT_FONT_STYLE; None to leave unconstrained
        :param font_size
            specify font size for table values - e.g. 6; None to leave unconstrained
        :param cell_padding
            specify left and right padding for table values - e.g. 5; None to leave unconstrained

        :returns builder
        """
        # add font style to table
        self.add_formatter(
            FormatterType.table,
            pbtf.FmtFontFamily(font_style, apply_to_header_and_index=False) if font_style is not None else None,
        )
        # add font size to table
        self.add_formatter(
            FormatterType.table,
            pbtf.FmtFontsize(font_size, unit="pt", apply_to_header_and_index=False) if font_size is not None else None,
        )
        # add column size to table
        self.add_formatter(
            FormatterType.table,
            pbtf.FmtHeader(
                fixed_width=table_width,
                column_width=column_width,
                no_wrap=False,
                columns=self._data.columns.tolist() if table_width is not None and column_width is not None else None,
            ),
        )
        # add padding to table
        self.add_formatter(
            FormatterType.table,
            pbtf.FmtAddCellPadding(
                left=cell_padding, right=cell_padding, top=0, bottom=0, apply_to_header_and_index=False
            )
            if cell_padding is not None
            else None,
        )
        # align table centre
        self.add_formatter(
            FormatterType.table,
            pbtf.FmtAlignTable(alignment=table_align, apply_to_header_and_index=False, columns=[pbtf.INDEX_COL_NAME])
            if table_align is not None
            else None,
        )
        # align values
        self.add_formatter(
            FormatterType.table,
            pbtf.FmtAlignCellContents(
                alignment=align_column, apply_to_header_and_index=False, columns=self._data.columns.tolist()
            )
            if align_column is not None
            else None,
        )
        return self

    def format_header(
        self,
        align_header: Optional[str],
        header_divider: bool,
        font_style: Optional[str],
        font_size: Optional[int],
        cell_padding: Optional[int],
    ) -> "CommonTableFormatterBuilder":
        """
        Set formatters for the header of the table

        :param align_header
            align header values - e.g. 'center', 'right' or 'left'; None to leave unconstrained
        :param header_divider
            include a dividing line between header row and body of the table
        :param font_style
            specify font style for header values - e.g. rs.DEFAULT_FONT_STYLE; None to leave unconstrained
        :param font_size
            specify font size for header values - e.g. 6; None to leave unconstrained
        :param cell_padding
            specify left and right padding for header values - e.g. 5; None to leave unconstrained

        :returns builder
        """
        # font style for header
        self.add_formatter(
            FormatterType.header,
            pbtf.FmtFontFamily(font_style, rows=[pbtf.HEADER_ROW_NAME], apply_to_header_and_index=False)
            if font_style is not None
            else None,
        )
        # font size for header
        self.add_formatter(
            FormatterType.header,
            pbtf.FmtFontsize(
                font_size,
                unit="pt",
                rows=[pbtf.HEADER_ROW_NAME],
                columns=[pbtf.INDEX_COL_NAME, *self._data.columns.tolist()],
                apply_to_header_and_index=False,
            )
            if font_size is not None
            else None,
        )
        # add padding to headers
        self.add_formatter(
            FormatterType.header,
            pbtf.FmtAddCellPadding(
                rows=[pbtf.HEADER_ROW_NAME],
                columns=[pbtf.INDEX_COL_NAME, *self._data.columns.tolist()],
                left=cell_padding,
                right=cell_padding,
                top=cell_padding,
                bottom=int(cell_padding / 2),
                apply_to_header_and_index=False,
            )
            if cell_padding is not None
            else None,
        )
        # align header
        self.add_formatter(
            FormatterType.header,
            pbtf.FmtAlignCellContents(
                alignment=align_header,
                rows=[pbtf.HEADER_ROW_NAME],
                columns=self._data.columns.tolist(),
                apply_to_header_and_index=False,
            )
            if align_header is not None
            else None,
        )
        # add line below header
        if header_divider:
            self.divider_line_horizontal(row=self._data.index[0], include_index=True)
        return self

    def format_index(
        self,
        index_width: Optional[str],
        align_index: Optional[str],
        index_bold: bool,
        font_style: Optional[str],
        font_size: Optional[int],
        cell_padding: Optional[int],
        table_width: Optional[str],
    ) -> "CommonTableFormatterBuilder":
        """
        Set formatters for the header of the table

        :param index_width
            specify column width as pct of table - e.g. '15%'; None to leave unconstrained
        :param align_index
            align index values - e.g. 'center', 'right' or 'left'; None to leave unconstrained
        :param index_bold
            whether the index values should be rendered bold
        :param font_style
            specify font style for index values - e.g. rs.DEFAULT_FONT_STYLE; None to leave unconstrained
        :param font_size
            specify font size for index values - e.g. 6; None to leave unconstrained
        :param cell_padding
            specify left and right padding for index values - e.g. 5; None to leave unconstrained
        :param table_width
            constrain width of table - e.g. '98%'; None to leave unconstrained

        :returns builder
        """
        # font style for index
        self.add_formatter(
            FormatterType.index,
            pbtf.FmtFontFamily(font_style, columns=[pbtf.INDEX_COL_NAME], apply_to_header_and_index=False)
            if font_style is not None
            else None,
        )
        # font size for index
        self.add_formatter(
            FormatterType.index,
            pbtf.FmtFontsize(font_size, unit="pt", columns=[pbtf.INDEX_COL_NAME], apply_to_header_and_index=False)
            if font_size is not None
            else None,
        )
        # width for index
        self.add_formatter(
            FormatterType.index,
            pbtf.FmtHeader(
                fixed_width=table_width, column_width=index_width, no_wrap=False, columns=[pbtf.INDEX_COL_NAME]
            )
            if table_width is not None and index_width is not None
            else None,
        )
        # add padding to index
        self.add_formatter(
            FormatterType.index,
            pbtf.FmtAddCellPadding(
                columns=[pbtf.INDEX_COL_NAME],
                left=cell_padding,
                right=cell_padding,
                top=0,
                bottom=0,
                apply_to_header_and_index=False,
            )
            if cell_padding is not None
            else None,
        )
        # align index
        self.add_formatter(
            FormatterType.index,
            pbtf.FmtAlignCellContents(
                alignment=align_index,
                apply_to_header_and_index=False,
                rows=[*self._data.index.tolist(), pbtf.HEADER_ROW_NAME],
                columns=[pbtf.INDEX_COL_NAME],
            )
            if align_index is not None
            else None,
        )
        # embolden index
        if index_bold:
            self.add_formatter(
                FormatterType.table, pbtf.FmtBold(columns=[pbtf.INDEX_COL_NAME], apply_to_header_and_index=True)
            )
        return self

    def truncate_contents_with_ellipsis(self, columns: Optional[List[str]] = None) -> "CommonTableFormatterBuilder":
        """
        Shorten values in specified columns to fit, if done then there will be a trailing ellipses

        :param columns
            columns to apply truncation formatting - None will cause formatting to be applied to all values

        :returns builder
        """
        self.add_formatter(
            FormatterType.truncate_contents_with_ellipsis, pbtf.FmtTruncateContentsWithEllipsis(columns=columns)
        )
        return self

    def replace_nans(self, value: str = "", columns: Optional[List[str]] = None) -> "CommonTableFormatterBuilder":
        """
        Replace NaNs in table

        :param value
             replace nan values with specified string - e.g. blank
        :param columns
            columns to apply NaN formatting - None will cause formatting to be applied to all values

        :returns builder
        """
        self.add_formatter(FormatterType.replace_nans, pbtf.FmtReplaceNaN(value=value, columns=columns))
        return self

    def decimal_columns(
        self, columns: Optional[List[str]], num_decimal_places: int = 2
    ) -> "CommonTableFormatterBuilder":
        """
        Apply decimal formatting to column values

        :param columns
            columns to apply decimal formatting - None will cause formatting to be applied to all columns
        :param num_decimal_places
            round to n decimal places

        :returns builder
        """
        self.add_formatter(
            FormatterType.decimal,
            pbtf.FmtDecimals(n=num_decimal_places, columns=columns, apply_to_header_and_index=False),
        )
        return self

    def bps_columns(
        self, columns: Optional[List[str]], num_decimal_places: int = 0, suffix: Optional[str] = " bps"
    ) -> "CommonTableFormatterBuilder":
        """
        Apply basis point formatting to column values

        :param columns
            columns to apply bps formatting - None will cause formatting to be applied to all columns
        :param num_decimal_places
            round to n decimal places
        :param suffix
            value to append to bps - None will prevent suffix from being added

        :returns builder
        """
        suffix = suffix if suffix is not None else ""
        self.add_formatter(
            FormatterType.bps, pbtf.FmtValueToBps(suffix=suffix, columns=columns, apply_to_header_and_index=False)
        )
        self.add_formatter(
            FormatterType.bps, pbtf.FmtDecimals(n=num_decimal_places, columns=columns, apply_to_header_and_index=False)
        )
        return self

    def pct_columns(
        self,
        columns: Optional[List[str]],
        num_decimal_places: int = 2,
        already_pct: bool = False,
        append_pct_sign: bool = False,
    ) -> "CommonTableFormatterBuilder":
        """
        Apply percentage formatting to column values

        :param columns
            columns to apply pct formatting - None will cause formatting to be applied to all columns
        :param num_decimal_places
            round to n decimal places
        :param already_pct
            no need to multiply by one hundred
        :param append_pct_sign
            add '%' as suffix to display value

        :returns builder
        """
        if not already_pct:
            self.add_formatter(
                FormatterType.percent, pbtf.FmtValueToPercent(columns=columns, apply_to_header_and_index=False)
            )
        if append_pct_sign:
            self.add_formatter(
                FormatterType.percent,
                pbtf.FmtPercent(n_decimals=num_decimal_places, columns=columns, apply_to_header_and_index=False),
            )
        else:
            self.add_formatter(
                FormatterType.percent,
                pbtf.FmtDecimals(n=num_decimal_places, columns=columns, apply_to_header_and_index=False),
            )
        return self

    def int_columns(self, columns: Optional[List[str]]) -> "CommonTableFormatterBuilder":
        """
        Apply integer formatting to column values

        :param columns
            columns to apply pct formatting - None will cause formatting to be applied to all columns

        :returns builder
        """
        self.add_formatter(FormatterType.int, pbtf.FmtDecimals(0, columns=columns, apply_to_header_and_index=False))
        return self

    def dollar_columns(self, columns: Optional[List[str]]) -> "CommonTableFormatterBuilder":
        """
        Apply dollar formatting to column values

        :param columns
            columns to apply dollar formatting - None will cause formatting to be applied to all columns

        :returns builder
        """
        self.add_formatter(
            FormatterType.dollar, pbtf.FmtThousandSeparator(columns=columns, apply_to_header_and_index=False)
        )
        return self

    def date_columns(self, columns: Optional[List[str]], dt_fmt: str = "{:%Y-%m-%d}") -> "CommonTableFormatterBuilder":
        """
        Apply date formatting to column values

        :param columns
            columns to apply date formatting - None will cause formatting to be applied to all columns
        :param dt_fmt
            date format string to be used

        :returns builder
        """
        self.add_formatter(
            FormatterType.date, pbtf.FmtDates(fmt_string=dt_fmt, columns=columns, apply_to_header_and_index=False)
        )
        return self

    def heatmap(
        self,
        rows: Optional[List[str]] = None,
        columns: Optional[List[str]] = None,
        axis: Optional[int] = None,
        center: float = 0.0,
        threshold: float = 0.0,
        max_color: Tuple[float, float, float] = colors.HEATMAP_RED,
        min_color: Tuple[float, float, float] = colors.HEATMAP_GREEN,
        center_color: Tuple[float, float, float] = colors.WHITE,
    ) -> "CommonTableFormatterBuilder":
        """
        Apply a heatmap over the values in the rows and columns specified - to apply row or columnwise specify axis

        :param rows
            rows to be included in the heatmap - None includes all
        :param columns
            columns to be included in the heatmap - None includes all
        :param axis
            specify row (0) or column (1) constraints to create multiple heatmaps - None means
            all values are included in one heatmap
        :param center The value to consider the centre of the distribution
        :param threshold bidirectional distance around center still considered the center
        :param max_color hot color
        :param min_color cold color
        :param center_color center color

        :returns builder
        """
        self.add_formatter(
            FormatterType.heat_map,
            pbtf.FmtHeatmapWithCenter(
                rows=rows,
                columns=columns,
                axis=axis,
                threshold=threshold,
                center=center,
                max_color=max_color,
                min_color=min_color,
                center_color=center_color,
            ),
        )
        return self

    def threshold(self, column: str, threshold_column: str) -> "CommonTableFormatterBuilder":
        """
        Highlight values in bold and italic if value exceeds threshold

        :param column
            column to use as source values and to apply formatting to
        :param threshold_column
            column to use to specify value that, if exceeded, causes the highlight to trigger

        :returns builder
        """
        self.add_formatter(
            FormatterType.threshold,
            pbtf.FmtHighlightText(
                rows=self._data.loc[self._data[column] > self._data[threshold_column]].index,
                columns=[column],
                bold=True,
                italic=True,
            ),
        )
        return self

    def highlight(self, columns: Optional[List[str]], highlight_column: str) -> "CommonTableFormatterBuilder":
        """
        Highlight values in bold and italic if highlight column is True

        :param columns
            columns to be formatted with conditional highlighting
        :param highlight_column
            source column of booleans that is used as the condition

        :returns builder
        """
        self.add_formatter(
            FormatterType.threshold,
            pbtf.FmtHighlightText(
                rows=self._data.loc[self._data[highlight_column]].index, columns=columns, bold=True, italic=True
            ),
        )
        return self

    def highlight_row(self, highlight_column: str) -> "CommonTableFormatterBuilder":
        """
        Highlight row in bold and italic if highlight column is True

        :param highlight_column
            source column of booleans that is used as the condition

        :returns builder
        """
        return self.highlight(
            columns=[pbtf.INDEX_COL_NAME, *self._data.columns.tolist()], highlight_column=highlight_column
        )

    def color_background(
        self,
        color: Tuple[float, float, float] = colors.RED,
        rows: Optional[List[str]] = None,
        columns: Optional[List[str]] = None,
    ) -> "CommonTableFormatterBuilder":
        """
        Colour the cell background

        :param color
            colour to be used as cell background
        :param rows
            rows to apply cell background colouring
        :param columns
            columns to apply cell background colouring

        :returns builder
        """
        self.add_formatter(
            FormatterType.color_background,
            pbtf.FmtHighlightBackground(color=color, rows=rows, columns=columns, apply_to_header_and_index=False),
        )
        return self

    def color_background_conditionally(
        self,
        condition: Callable,
        color: Tuple[float, float, float] = colors.RED,
        rows: Optional[List[str]] = None,
        columns: Optional[List[str]] = None,
    ) -> "CommonTableFormatterBuilder":
        """
        Colour the cell background conditionally

        :param condition
            condition to apply format, should be function that will take single cell value and return bool
        :param color
            colour to be used as cell background
        :param rows
            rows to apply cell background colouring
        :param columns
            columns to apply cell background colouring

        :returns builder
        """
        rows = rows or self._data.index
        columns = columns or self._data.columns
        for column in columns:
            conditional_data = self._data.loc[rows][column]
            conditional_rows = conditional_data.loc[conditional_data.apply(condition)].index
            self.color_background(color=color, rows=conditional_rows, columns=[column])
        return self

    def color_background_conditionally_matching(
        self,
        value: Union[str, float, int],
        color: Tuple[float, float, float] = colors.RED,
        rows: Optional[List[str]] = None,
        columns: Optional[List[str]] = None,
    ) -> "CommonTableFormatterBuilder":
        """
        Colour the cell background if value matches

        :param value
            value to match
        :param color
            colour to be used as cell background
        :param rows
            rows to apply cell background colouring
        :param columns
            columns to apply cell background colouring

        :returns builder
        """
        self.color_background_conditionally(condition=lambda v: v == value, color=color, rows=rows, columns=columns)
        return self

    def divider_line_vertical(
        self, column: str, include_header: bool = False, color: Tuple[float, float, float] = colors.BLUE
    ) -> "CommonTableFormatterBuilder":
        """
        Insert a vertical line into table

        :param column
            column to add dividing line to the left of
        :param include_header
            should dividing line extend through table header
        :param color
            colour of line - defaults to blue

        :returns builder
        """
        self.add_formatter(
            FormatterType.divider_lines,
            pbtf.FmtAddCellBorder(
                rows=([pbtf.HEADER_ROW_NAME] if include_header else []) + self._data.index.tolist(),
                columns=[column],
                left=1,
                color=color,
                apply_to_header_and_index=False,
            ),
        )
        return self

    def divider_line_horizontal(
        self, row: str, include_index: bool = False, color=colors.BLUE
    ) -> "CommonTableFormatterBuilder":
        """
        Insert a horizontal line into table

        :param row
            row to add dividing line above
        :param include_index
            should dividing line extend through table index
        :param color
            colour of line - defaults to blue

        :returns builder
        """
        self.add_formatter(
            FormatterType.divider_lines,
            pbtf.FmtAddCellBorder(
                rows=[row],
                columns=([pbtf.INDEX_COL_NAME] if include_index else []) + self._data.columns.tolist(),
                top=1,
                color=color,
                apply_to_header_and_index=False,
            ),
        )
        return self

    def hide_columns(self, columns: List[str]) -> "CommonTableFormatterBuilder":
        """
        Hide specified columns from the rendered table - they are still available for conditional formatters
            such as  highlight, threshold and heatmap

        :param columns
            columns to hide

        :returns builder
        """
        self.add_formatter(FormatterType.hide_cells, pbtf.FmtHideCells(columns=columns))
        return self

    def hide_index(self) -> "CommonTableFormatterBuilder":
        """
        Hide index from the rendered table

        :returns builder
        """
        self.hide_columns(columns=[pbtf.INDEX_COL_NAME])
        return self

    def hide_insignificant(self, columns: Optional[List[str]] = None) -> "CommonTableFormatterBuilder":
        """
        Replace zero values with blank strings in columns specified

        :param columns
            columns to be formatted with conditional blanking - None will cause formatting to be applied to all columns

        :returns builder
        """
        self.add_formatter(
            FormatterType.hide_insignificant,
            pbtf.FmtHideInsignificant(columns=columns, apply_to_header_and_index=False),
        )
        return self

    def stripe_background(self) -> "CommonTableFormatterBuilder":
        """
        Stripe background of table rows in grey and white

        :returns builder
        """
        self.add_formatter(FormatterType.stripe_background, pbtf.FmtStripeBackground())
        return self

    def total(
        self, columns: Optional[List[str]], total_row_name="Total", total_divider: bool = True
    ) -> "CommonTableFormatterBuilder":
        """
        Add a total row to the bottom of the table, performing a simple sum of values in column.

        :param columns
            columns to be totalised - None indicates all columns possible
        :param total_row_name
            index label for total row
        :param total_divider
            specify whether to have a dividing line between table and total row

        :returns builder
        """
        self.add_formatter(
            FormatterType.total,
            pbtf.FmtAppendTotalsRow(
                row_name=total_row_name,
                total_columns=columns,
                hline_color=colors.BLUE if total_divider else None,
                hline_style="1px solid" if total_divider else None,
            ),
        )
        return self



from pybloqs.config import user_config
from pybloqs.htmlconv.chrome_headless import ChromeHeadlessConverter
from pybloqs.htmlconv.html_converter import HTMLConverter
from pybloqs.htmlconv.wkhtmltox import WkhtmltoimageConverter, WkhtmltopdfConverter

PDF_CONVERTERS = {
    "wkhtmltopdf": WkhtmltopdfConverter,
    "chrome_headless": ChromeHeadlessConverter,
}

IMAGE_CONVERTERS = {
    "wkhtmltoimage": WkhtmltoimageConverter,
}


def get_converter(file_type: str) -> HTMLConverter:
    """Parse the config and return the right backend for converting HTML to other formats."""
    if file_type.lower() == "pdf":
        converter_name = user_config["pdf_converter"]
        return PDF_CONVERTERS[converter_name]()
    elif file_type.lower() in ["png", "svg", "jpg"]:
        converter_name = user_config["image_converter"]
        return IMAGE_CONVERTERS[converter_name]()
    else:
        raise ValueError(f"No converter defined for file type: {file_type}")



from collections import namedtuple
from typing import Any, Iterable, List, Optional, Tuple

import pandas as pd
from jinja2 import Environment, PackageLoader

from pybloqs.block.base import BaseBlock
from pybloqs.block.convenience import add_block_types
from pybloqs.block.table_formatters import DEFAULT_DECIMALS_FORMATTER, DEFAULT_FORMATTERS, ORG_ROW_NAMES, TableFormatter
from pybloqs.html import parse

_jinja_env = Environment(loader=PackageLoader("pybloqs", "jinja"))
_jinja_env.globals.update(len=len)
_jinja_env.globals.update(enumerate=enumerate)
_jinja_env.globals.update(slice=slice)
_jinja_env.globals.update(zip=zip)
_table_tmpl = _jinja_env.get_template("table.html")


IndexCell = namedtuple("IndexCell", ["value", "names", "span", "depth"])


class HTMLJinjaTableBlock(BaseBlock):
    FormatterData = namedtuple("FormatterData", ["cell", "row_name", "column_name", "df"])

    def __init__(
        self,
        df: pd.DataFrame,
        formatters: Optional[List[TableFormatter]] = None,
        use_default_formatters: bool = True,
        merge_vertical: bool = False,
        **kwargs,
    ) -> None:
        """Create table from Jinja framework. Apply formatters to customise table formatting.

        Parameters()
        ----------
        df: 'DataFrame'
            The DF from which the values are taken.
        formatters: 'list'
            List of formatters, which are objects inheriting from TableFormatter class.
        """
        super().__init__(**kwargs)
        if formatters is None:
            formatters = []
        if use_default_formatters:
            formatters = DEFAULT_FORMATTERS + formatters + DEFAULT_DECIMALS_FORMATTER

        self.formatters = formatters
        # Apply modifications to DataFrame at the earliest stage.
        for formatter in formatters:
            try:
                df = formatter.modify_dataframe(df)
            except NotImplementedError:
                continue
        self.df = df
        self.n_header_rows = len(df.columns.names)
        self.merge_vertical = merge_vertical

    def modify_cell_content(self, cell, row_name, column_name) -> Any:
        if ORG_ROW_NAMES in self.df.columns and self.row_index > 0:
            row_name = self.df[ORG_ROW_NAMES].iloc[self.row_index]
        for formatter in self.formatters:
            try:
                data = self.FormatterData(cell, row_name, column_name, self.df)
                cell = formatter.modify_cell_content(data)
            except NotImplementedError:
                continue
        return cell

    def insert_additional_html(self) -> str:
        html_string = ""
        for formatter in self.formatters:
            try:
                html_string += formatter.insert_additional_html()
            except NotImplementedError:
                continue
        return html_string

    def _join_css_substrings(self, css_substrings: Iterable[str], prefix: str) -> str:
        return prefix + '="' + "; ".join(css_substrings) + '"'

    def _aggregate_css_formatters(
        self, function_name: str, fmt_args: Optional[List[Any]] = None, prefix: str = "style"
    ) -> str:
        css_substrings = []
        fmt_args = fmt_args if fmt_args else []
        for formatter in self.formatters:
            try:
                fmt_func = getattr(formatter, function_name)
                css_substring = fmt_func(*fmt_args)
            except NotImplementedError:
                continue
            if css_substring is not None:
                css_substrings.append(css_substring)
        return self._join_css_substrings(css_substrings, prefix)

    def create_table_level_css(self) -> str:
        self.row_index = -self.n_header_rows - 1
        return self._aggregate_css_formatters("create_table_level_css")

    def create_table_level_css_class(self) -> str:
        return self._aggregate_css_formatters("create_table_level_css_class", prefix="class")

    def create_thead_level_css(self) -> str:
        return self._aggregate_css_formatters("create_thead_level_css")

    def create_row_level_css(self, row_name: str, row) -> str:
        self.row_index += 1
        if ORG_ROW_NAMES in self.df.columns and self.row_index >= 0:
            row_name = self.df[ORG_ROW_NAMES].iloc[self.row_index]
        data = pd.Series(row, name=row_name)
        return self._aggregate_css_formatters("create_row_level_css", fmt_args=[data])

    def create_column_level_css(self, column_name: str, series) -> str:
        data = self.FormatterData(None, None, column_name, series)
        return self._aggregate_css_formatters("create_column_level_css", fmt_args=[data])

    def create_cell_level_css(self, cell, row_name: str, column_name: str) -> str:
        if ORG_ROW_NAMES in self.df.columns and self.row_index >= 0:
            row_name = self.df[ORG_ROW_NAMES].iloc[self.row_index]
        data = self.FormatterData(cell, row_name, column_name, self.df)
        return self._aggregate_css_formatters("create_cell_level_css", fmt_args=[data])

    def _get_header_iterable(self) -> List[List[IndexCell]]:
        df_clean = self.df.loc[:, self.df.columns.get_level_values(0) != ORG_ROW_NAMES]
        return columns_to_iterable(df_clean.columns, merge_depth=self.merge_vertical)

    def _get_index_iterable(self) -> List[List[IndexCell]]:
        return index_to_iterable(self.df.index)

    def _write_contents(self, container, actual_cfg, *args, **kwargs) -> None:
        # table boilerplate
        model = {
            "df": self.df,
            "header_iterable": self._get_header_iterable(),
            "index_iterable": self._get_index_iterable(),
            "insert_additional_html": self.insert_additional_html,
            "create_thead_level_css": self.create_thead_level_css,
            "create_table_level_css": self.create_table_level_css,
            "create_table_level_css_class": self.create_table_level_css_class,
            "create_column_level_css": self.create_column_level_css,
            "create_row_level_css": self.create_row_level_css,
            "create_cell_level_css": self.create_cell_level_css,
            "modify_cell_content": self.modify_cell_content,
        }

        table_html = _table_tmpl.render(**model)
        soup = parse(table_html)
        table = soup.find("table")
        container.append(table)


def multiindex_to_tuples(index: pd.Index) -> List[Tuple]:
    return [tuple(col) for col in index]


def index_to_iterable(index: pd.Index, merge_depth: bool = False) -> List[List[IndexCell]]:
    """
    Return the given index as a list of lists of (potentially merged) cells
    suitable for rendering as HTML, in span-major order. i.e. suited for
    rendering a table index.

    Each cell is an IndexCell namedtuple representing one <td> tag with:
    * `value`: the content of that cell
    * `names`: a list of index values over the span of this cell
    * `span`: the number of index values covered by this cell (rowspan if index, colspan if header)
    * `depth`: the number of MultiIndex levels covered by this cell (colspan if index, rowspan if header)
    """
    sentinel = object()
    if isinstance(index, pd.MultiIndex):
        num_levels = len(index.names)

        # convert index to tuples and reverse the order to help the merging
        # logic below.
        values = index.tolist()
        values = list(reversed(values))

        values.append((sentinel,) * num_levels)
        carryover = [0] * num_levels

        result = []
        for rownum, (row, prev_row) in enumerate(zip(values, values[1:])):
            # find first cell which does not match the cell before it in the
            # index. only until num_levels-1 because the deepest cell should
            # never be merged.
            for depth in range(0, num_levels - 1):
                if row[depth] == prev_row[depth]:
                    # keep track of 'carryover': the number of cells that were
                    # omitted. this will be added to the span of the merged
                    # cell.
                    carryover[depth] += 1
                else:
                    break
            else:
                depth = num_levels - 1

            # generate entries for the remaining cells.
            cells = []
            for depth in range(depth, num_levels):
                cells.append(
                    IndexCell(
                        row[depth],
                        list(reversed(values[rownum - carryover[depth] : rownum + 1])),
                        carryover[depth] + 1,
                        1,
                    )
                )
                carryover[depth] = 0

            # merge cells depth-wise (i.e. vertically if header, horizontally
            # if index) if required.
            if merge_depth:
                merged_cells = []
                merge_count = 0
                for cell, next_cell in zip(cells[:-1], cells[1:-1] + [IndexCell(sentinel, [], 0, 0)]):
                    merge_count += 1
                    if cell.value == next_cell.value and cell.span == next_cell.span:
                        continue
                    merged_cells.append(IndexCell(cell.value, cell.names, cell.span, merge_count))
                    merge_count = 0
                cells = [*merged_cells, cells[-1]]

            result.append(cells)

        return list(reversed(result))
    else:
        return [[IndexCell(value, [value], 1, 1)] for value in index.tolist()]


def columns_to_iterable(column_index: pd.Index, merge_depth: bool = False) -> List[List[IndexCell]]:
    """
    Return the given index as a list of lists of (potentially merged) cells
    suitable for rendering as HTML, in depth-major order. i.e. suited for
    rendering a table header.
    """
    rows = index_to_iterable(column_index, merge_depth=merge_depth)
    result = [[] for _ in range(len(column_index.names))]

    # transpose the index iterable from span-major to depth-major order.
    skips = [0] * len(column_index.names)
    for distance, row in enumerate(rows):
        # if a previous cell at this depth had a span that overlaps, then don't
        # emit a cell at this depth.
        depth = 0
        while skips[depth]:
            skips[depth] -= 1
            depth += 1

        for cell in row:
            result[depth].append(cell)
            for skip_index in range(depth, depth + cell.depth):
                skips[skip_index] = cell.span - 1
            depth += cell.depth

    return result


add_block_types(pd.DataFrame, HTMLJinjaTableBlock)



import os
from numbers import Number
from typing import TYPE_CHECKING, Optional, Tuple, Union

from pybloqs.htmlconv.html_converter import A4, PORTRAIT, HTMLConverter

if TYPE_CHECKING:
    from pybloqs import BaseBlock


class WkhtmltopdfConverter(HTMLConverter):
    """
    Implementation of HTML to PDF conversion using wkhtmltopdf tool.
    """

    def htmlconv(
        self,
        block: "BaseBlock",
        output_file: str,
        header_block: Optional["BaseBlock"] = None,
        header_spacing: Optional[Union[str, float]] = None,
        footer_block: Optional["BaseBlock"] = None,
        footer_spacing: Optional[Union[str, float]] = None,
        pdf_zoom: Number = 1,
        pdf_page_size: str = A4,
        orientation: str = PORTRAIT,
        pdf_auto_shrink: bool = True,
        **kwargs: str,
    ) -> Tuple[bytes, bytes]:
        """
        Use the wkhtmltohtml tool to convert the supplied HTML content to a PDF file.

        :param block: The block to render
        :param output_file: File name for converter output.
        :param header_block: A header block to add
        :param header_spacing: The spacing for the header as number in mm.
        :param footer_block: A footer block to add
        :param footer_spacing: The spacing for the footer as number in mm.
        :param pdf_zoom: The zooming to apply when rendering the page.
        :param pdf_page_size: The page size to use when rendering the page to PDF.
        :param orientation: Either html_converter.PORTRAIT or html_converter.LANDSCAPE
        :param kwargs: Additional parameters. Passed to wkhtmltopdf.
        """
        cmd = []
        cmd.append(self.get_executable("wkhtmltopdf"))
        cmd += ["--page-size", pdf_page_size]
        cmd += ["--orientation", orientation]
        cmd += ["--zoom", str(pdf_zoom)]
        if pdf_auto_shrink:
            cmd.append("--enable-smart-shrinking")
        else:
            cmd.append("--disable-smart-shrinking")

        # Specific wkhtmltox
        cmd += ["--no-stop-slow-scripts", "--debug-javascript"]
        cmd += ["--javascript-delay", str(kwargs.get("javascript_delay", 200))]
        [cmd.extend([k, v]) for k, v in kwargs.items()]

        # Render without header and footer as those are handles explicitly below
        content = block.render_html(static_output=True, permit_compression=False)
        html_filename = HTMLConverter.write_html_to_tempfile(block, content)

        temp_files = [html_filename]

        if header_block is not None:
            header_content = header_block.render_html(static_output=True, permit_compression=False)
            header_filename = HTMLConverter.write_html_to_tempfile(header_block, header_content)
            cmd += ["--header-html", header_filename]
            cmd += ["--header-spacing", str(header_spacing)]
            temp_files.append(header_filename)

        if footer_block is not None:
            footer_content = footer_block.render_html(static_output=True, permit_compression=False)
            footer_filename = HTMLConverter.write_html_to_tempfile(footer_block, footer_content)
            cmd += ["--footer-html", footer_filename]
            cmd += ["--footer-spacing", str(footer_spacing)]
            temp_files.append(footer_filename)

        cmd.extend([html_filename, output_file])

        output, errors = self.run_command(cmd)
        HTMLConverter.remove_temporary_files(temp_files)
        return output, errors


class WkhtmltoimageConverter(HTMLConverter):
    def htmlconv(self, block: "BaseBlock", output_file: str, pdf_zoom: float = 1, **kwargs) -> Tuple[bytes, bytes]:
        """
        Use the wkhtmltoimage tool to convert the supplied HTML content to an image file.

        :param block: The block to render
        :param output_file: File name for converter output.
        :param pdf_zoom: The zooming to apply when rendering the page.
        :param kwargs: Additional parameters. Passed to wkhtmltoimage.
        """
        content = block.render_html(static_output=True, permit_compression=False)
        html_filename = HTMLConverter.write_html_to_tempfile(block, content)

        cmd = []
        cmd.append(self.get_executable("wkhtmltoimage"))
        cmd += ["--format", os.path.splitext(output_file)[1][1:]]
        cmd += ["--zoom", str(pdf_zoom)]
        # Set width to 0 to make it fit the content exactly, without any extra margin.
        cmd += ["--width", "0"]

        # Specific wkhtmltox
        cmd += ["--no-stop-slow-scripts", "--debug-javascript"]
        cmd += ["--javascript-delay", str(kwargs.get("javascript_delay", 200))]

        # Ignore default parameters that are not relevant for image output and pass the rest to wkhtmltoimage
        kwargs_ignore_list = [
            "header_block",
            "header_spacing",
            "footer_block",
            "footer_spacing",
            "pdf_auto_shrink",
            "pdf_page_size",
            "orientation",
        ]
        [cmd.extend([k, v]) for k, v in kwargs.items() if k not in kwargs_ignore_list]

        cmd.extend([html_filename, output_file])

        output, errors = self.run_command(cmd)
        HTMLConverter.remove_temporary_files([html_filename])
        return output, errors



from pybloqs.block.base import BaseBlock, HRule
from pybloqs.block.convenience import Block
from pybloqs.block.image import ImgBlock, PlotBlock, set_plot_format
from pybloqs.block.layout import Flow, Grid, HStack, VStack
from pybloqs.block.table import HTMLJinjaTableBlock
from pybloqs.block.text import Markdown, Pre, Raw, Span
from pybloqs.block.wrap import Box, Paragraph
from pybloqs.util import Cfg

try:
    from importlib.metadata import PackageNotFoundError, distribution

    try:
        dist = distribution(__package__)
        __version__ = dist.version
    except PackageNotFoundError:
        __version__ = "<unknown>"
except ImportError:
    from pkg_resources import DistributionNotFound, get_distribution

    try:
        __version__ = get_distribution(__name__).version
    except DistributionNotFound:
        __version__ = "<unknown>"

__all__ = [
    # Core PyBloqs
    "set_plot_format",
    "Block",
    "BaseBlock",
    "HRule",
    "Raw",
    "Grid",
    "Flow",
    "HStack",
    "VStack",
    "ImgBlock",
    "PlotBlock",
    "HTMLJinjaTableBlock",
    "Box",
    "Paragraph",
    "Pre",
    "Span",
    "Markdown",
    "Cfg",
]
